# -*- coding: utf-8 -*-

"""COST FUNCTIONS

This module contains classes of different cost functions for optimization.

:Author: Samuel Farrens <samuel.farrens@gmail.com>

:Version: 2.2

:Date: 23/10/2017

"""

from __future__ import division, print_function
import numpy as np
try:
    from sf_tools.plot.cost_plot import plotCost
except ImportError:
    import_fail = True
else:
    import_fail = False


class costObj(object):

    """Generic cost function object

    This class updates the cost according to the input cost functio class and
    tests for convergence

    Parameters
    ----------
    costFunc : class
        Class for calculating the cost
    initial_cost : float, optional
        Initial value of the cost (default is "1e6")
    tolerance : float, optional
        Tolerance threshold for convergence (default is "1e-4")
    cost_interval : int, optional
        Iteration interval to calculate cost (default is "1")
    test_range : int, optional
        Number of cost values to be used in test (default is "4")
    verbose : bool, optional
        Option for verbose output (default is "True")
    plot_output : str, optional
        Output file name for cost function plot

    Notes
    -----
    The costFunc class must contain a method called `calc_cost()`.

    """

    def __init__(self, costFunc, initial_cost=1e6, tolerance=1e-4,
                 cost_interval=1, test_range=4, verbose=True,
                 plot_output=None):

        if not hasattr(costFunc, 'calc_cost'):
            raise ValueError('costFunc must contain "calc_cost" method.')

        self.costFunc = costFunc
        self.cost = initial_cost
        self._cost_list = []
        self._cost_interval = cost_interval
        self._iteration = 1
        self._test_list = []
        self._test_range = test_range
        self._tolerance = tolerance
        self._plot_output = plot_output
        self._verbose = verbose

    def _check_cost(self):
        """Check cost function

        This method tests the cost function for convergence in the specified
        interval of iterations using the last n (test_range) cost values

        Returns
        -------
        bool result of the convergence test

        """

        # Add current cost value to the test list
        self._test_list.append(self.cost)

        # Check if enough cost values have been collected
        if len(self._test_list) == self._test_range:

            # The mean of the first half of the test list
            t1 = np.mean(self._test_list[len(self._test_list) // 2:], axis=0)
            # The mean of the second half of the test list
            t2 = np.mean(self._test_list[:len(self._test_list) // 2], axis=0)
            # Calculate the change across the test list
            cost_diff = (np.linalg.norm(t1 - t2) / np.linalg.norm(t1))
            # Reset the test list
            self._test_list = []

            if self._verbose:
                print(' - CONVERGENCE TEST - ')
                print(' - CHANGE IN COST:', cost_diff)
                print('')

            # Check for convergence
            return cost_diff <= self._tolerance

        else:

            return False

    def get_cost(self, *args, **kwargs):
        """Get cost function

        This method calculates the current cost and tests for convergence

        Returns
        -------
        bool result of the convergence test

        """

        # Check if the cost should be calculated
        if self._iteration % self._cost_interval:

            test_result = False

        else:

            if self._verbose:
                print(' - ITERATION:', self._iteration)

            # Calculate the current cost
            self.cost = self.costFunc.calc_cost(*args, **kwargs)
            self._cost_list.append(self.cost)

            if self._verbose:
                print(' - Log10 COST:', np.log10(self.cost))
                print('')

            # Test for convergence
            test_result = self._check_cost()

        # Update the current iteration number
        self._iteration += 1

        return test_result

    def plot_cost(self):
        """Plot the cost function

        This method plots the cost function as function of iteration number

        """

        if import_fail:
            pass

        else:
            plotCost(self._cost_list, self._plot_output)



# -*- coding: utf-8 -*-

"""REGRESSION ROUTINES

This module contains methods for linear regression.

:Author: Samuel Farrens <samuel.farrens@gmail.com>

:Version: 1.0

:Date: 04/04/2017

"""

from builtins import zip
import numpy as np
from scipy.odr import *


def linear_fit(B, x):
    r"""Linear fit

    This method defines the equation of a straight line.

    Parameters
    ----------
    B : tuple
        Slope (m) and intercept (b) of the line.
    x : list or np.ndarray
        The 1D data vector

    Returns
    -------
    np.ndarray 1D array of corresponding y values

    NOTES
    -----
    This equation of a stright line is given by

    .. math::

        y = mx + b

    """

    return B[0] * np.array(x) + B[1]


def polynomial(x, a):
    r"""Polynomial

    This method defines the equation of a polynomial line.

    Parameters
    ----------
    x : list or np.ndarray
        The 1D data vector
    a : list or np.ndarray
        The 1D polynomial coefficient vector

    Returns
    -------
    np.ndarray 1D array of corresponding y values

    NOTES
    -----
    This equation of a stright line is given by

    .. math::

        y = a_0 + a_1x + a_2x^2 + \dots + a_kx^k

    """

    a = np.array(a)
    x = np.array(x)

    return sum([(a_i * x ** n) for a_i, n in zip(a, range(a.size))])


def polynomial_fit(x, y, k=1):
    """Polynomial fit

    This method finds the coefficients for a polynomial line fit to the input
    data using least squares.

    Parameters
    ----------
    x : list or np.ndarray
        The 1D independent data vector
    y : list or np.ndarray
        The 1D dependent data vector
    k : int, optional
        Number of degrees of freedom. Default (k=1)

    Returns
    -------
    np.ndarray 1D array of coefficients a

    """

    y = np.array(y)

    return least_squares(x_matrix(x, k), y)


def least_squares(X, y):
    """Least squares

    This method performs an analytical least squares regression. Returns the
    values of the coefficients, a, given the input matrix X and the
    corresponding y values.

    Parameters
    ----------
    X : np.ndarray
        The 2D independent data matrix.
    y : np.ndarray
        The 1D dependent data vector

    Returns
    -------
    np.ndarray 1D array of coefficients a

    Raises
    ------
    ValueError
        If inputs are not numpy arrays

    ToDo
    ----
    Add equation and example

    """

    if not np.all([isinstance(i, np.ndarray) for i in (X, y)]):
        raise ValueError('Inputs must be numpy arrays.')

    return np.dot(np.dot(np.linalg.inv(np.dot(X.T, X)), X.T), y)


def x_matrix(x, k):
    """Define X matrix

    This method defines the matrix X for a given vector x corresponding to a
    polynomial with k degrees of freedom.

    Parameters
    ----------
    x : list or np.ndarray
        The 1D independent data vector
    k : int
        Number of degrees of freedom

    Returns
    -------
    np.ndarray the 2D independent variable matrix X

    """

    x = np.array(x)

    return np.vstack([x ** n for n in range(k + 1)]).T


def fit_odr(x, y, xerr, yerr, fit):
    """Fit via ODR

    This method performs an orthogonal distance regression fit.

    Parameters
    ----------
    x : list or np.ndarray
        The 1D independent data vector
    y : list or np.ndarray
        The 1D dependent data vector
    x_err : list or np.ndarray
        1D data vector of x value errors
    y_err : list or np.ndarray
        1D data vector of y value errors
    fit : function
        Fitting function

    Returns
    -------
    tuple best fit parameters

    """

    model = Model(fit)
    r_data = RealData(x, y, sx=xerr, sy=yerr)
    odr = ODR(r_data, model, beta0=[1.0, 2.0])
    odr_out = odr.run()

    return odr_out.beta



# -*- coding: utf-8 -*-

"""ERROR HANDLING ROUTINES

This module contains methods for handing warnings and errors.

:Author: Samuel Farrens <samuel.farrens@gmail.com>

:Version: 1.2

:Date: 23/10/2017

"""


import sys
import os.path
import warnings
try:
    from termcolor import colored
except ImportError:
    import_fail = True
else:
    import_fail = False


def warn(warn_string, log=None):
    """Warning

    This method creates custom warning messages.

    Parameters
    ----------
    warn_string : str
        Warning message string
    log : instance, optional
        Logging structure instance

    """

    if import_fail:
        warn_txt = 'WARNING'
    else:
        warn_txt = colored('WARNING', 'yellow')

    # Print warning to stdout.
    sys.stderr.write(warn_txt + ': ' + warn_string + '\n')

    # Check if a logging structure is provided.
    if not isinstance(log, type(None)):
        warnings.warn(warn_string)


def catch_error(exception, log=None):
    """Catch error

    This method catches errors and prints them to the terminal. It also saves
    the errors to a log if provided.

    Parameters
    ----------
    exception : str
        Exception message string
    log : instance, optional
        Logging structure instance

    """

    if import_fail:
        err_txt = 'ERROR'
    else:
        err_txt = colored('ERROR', 'red')

    # Print exception to stdout.
    stream_txt = err_txt + ': ' + str(exception) + '\n'
    sys.stderr.write(stream_txt)

    # Check if a logging structure is provided.
    if not isinstance(log, type(None)):
        log_txt = 'ERROR: ' + str(exception) + '\n'
        log.exception(log_txt)


def file_name_error(file_name):
    """File name error

    This method checks if the input file name is valid.

    Parameters
    ----------
    file_name : str
        File name string

    Raises
    ------
    IOError
        If file name not specified or file not found

    """

    if file_name == '' or file_name[0][0] == '-':
        raise IOError('Input file name not specified.')

    elif not os.path.isfile(file_name):
        raise IOError('Input file name [%s] not found!' % file_name)



# -*- coding: utf-8 -*-

"""STATISTICS ROUTINES

This module contains methods for basic statistics.

:Author: Samuel Farrens <samuel.farrens@gmail.com>

:Version: 1.2

:Date: 20/10/2017

"""

from __future__ import division
from builtins import zip
import numpy as np
from scipy.stats import chi2
from astropy.convolution import Gaussian2DKernel


def chi2_gof(data_obs, data_exp, sigma, ddof=1):
    """Chi-squared goodness-of-fit

    This method tests the chi^2 goodness of fit.

    Parameters
    ----------
    data_obs : np.ndarray
        Observed data array
    data_exp : np.ndarray
        Expected data array
    sigma : float
        Expected data error
    ddof : input
        Delta degrees of freedom. Default (ddof = 1).

    Returns
    -------
    tuple of floats chi-squared and P values

    """

    chi2 = np.sum(((data_obs - data_exp) / sigma) ** 2)
    p_val = chi2.cdf(chi2, len(data_obs) - ddof)

    return chi2, p_val


def gaussian(point, mean, sigma, amplitude=None):
    """Gaussian distribution

    Method under development...

    """

    if isinstance(amplitude, type(None)):
        amplitude = 1

    val = np.array([((x - mu) / sig) ** 2 for x, mu, sig in
                   zip(point, mean, sigma)])

    return amplitude * np.exp(-0.5 * val)


def gaussian_kernel(data_shape, sigma, norm='max'):
    """Gaussian kernel

    This method produces a Gaussian kerenal of a specified size and dispersion

    Parameters
    ----------
    data_shape : tuple
        Desiered shape of the kernel
    sigma : float
        Standard deviation of the kernel
    norm : str {'max', 'sum'}, optional
        Normalisation of the kerenl (options are 'max' or 'sum')

    Returns
    -------
    np.ndarray kernel

    """

    kernel = np.array(Gaussian2DKernel(sigma, x_size=data_shape[1],
                      y_size=data_shape[0]))

    if norm is 'max':
        return kernel / np.max(kernel)

    elif norm is 'sum':
        return kernel / np.sum(kernel)

    else:
        return kernel


def mad(data):
    r"""Median absolute deviation

    This method calculates the median absolute deviation of the input data.

    Parameters
    ----------
    data : np.ndarray
        Input data array

    Returns
    -------
    float MAD value

    Notes
    -----
    The MAD is calculated as follows:

    .. math::

        \mathrm{MAD} = \mathrm{median}\left(|X_i - \mathrm{median}(X)|\right)

    """

    return np.median(np.abs(data - np.median(data)))


def mse(data1, data2):
    """Mean Squared Error

    This method returns the Mean Squared Error (MSE) between two data sets.

    Parameters
    ----------
    data1 : np.ndarray
        First data set
    data2 : np.ndarray
        Second data set

    """

    return np.mean((data1 - data2) ** 2)


def psnr2(image, noisy_image, max_pix=255):
    r"""Peak Signal-to-Noise Ratio

    This method calculates the PSNR between an image and a noisy version
    of that image

    Parameters
    ----------
    image : np.ndarray
        Input image, 2D array
    noisy_image : np.ndarray
        Noisy image, 2D array
    max_pix : int
        Maximum number of pixels. Default (max_pix=255)

    Returns
    -------
    float PSNR value

    Notes
    -----
    Implements PSNR equation on
    https://en.wikipedia.org/wiki/Peak_signal-to-noise_ratio

    .. math::

        \mathrm{PSNR} = 20\log_{10}(\mathrm{MAX}_I - 10\log_{10}(\mathrm{MSE}))

    """

    return (20 * np.log10(max_pix) - 10 *
            np.log10(mse(image, noisy_image)))


def psnr(image, recovered_image):
    """Peak Signal-to-Noise Ratio

    This method calculates the PSNR between an image and the recovered version
    of that image

    Parameters
    ----------
    image : np.ndarray
        Input image, 2D array
    recovered_image : np.ndarray
        Recovered image, 2D array

    Returns
    -------
    float PSNR value

    Notes
    -----
    Implements eq.3.7 from _[S2010]

    """

    return (20 * np.log10((image.shape[0] * np.abs(np.max(image) -
            np.min(image))) / np.linalg.norm(image - recovered_image)))


def psnr_stack(images, recoverd_images, metric=np.mean):
    """Peak Signa-to-Noise for stack of images

    This method calculates the PSNRs for a stack of images and the
    corresponding recovered images. By default the metod returns the mean
    value of the PSNRs, but any other metric can be used.

    Parameters
    ----------
    images : np.ndarray
        Stack of images, 3D array
    recovered_images : np.ndarray
        Stack of recovered images, 3D array
    metric : function
        The desired metric to be applied to the PSNR values (default is
        'np.mean')

    Returns
    -------
    float metric result of PSNR values

    Raises
    ------
    ValueError
        For invalid input data dimensions

    """

    if images.ndim != 3 or recoverd_images.ndim != 3:
        raise ValueError('Input data must be a 3D np.ndarray')

    return metric([psnr(i, j) for i, j in zip(images, recoverd_images)])


def sigma_mad(data):
    r"""Standard deviation from MAD

    This method calculates the standard deviation of the input data from the
    MAD.

    Parameters
    ----------
    data : np.ndarray
        Input data array

    Returns
    -------
    float sigma value

    Notes
    -----
    This function can be used for estimating the standeviation of the noise in
    imgaes.

    Sigma is calculated as follows:

    .. math::

        \sigma = 1.4826 \mathrm{MAD}(X)

    """

    return 1.4826 * mad(data)



# -*- coding: utf-8 -*-

"""FILTER ROUTINES

This module contains methods for distance measurements in cosmology.

:Author: Samuel Farrens <samuel.farrens@gmail.com>

:Version: 1.1

:Date: 20/10/2017

"""

from __future__ import division
import numpy as np
from sf_tools.base.types import check_float


def Gaussian_filter(x, sigma, norm=True):
    """Gaussian filter

    This method implements a Gaussian filter.

    Parameters
    ----------
    x : float
        Input data point
    sigma : float
        Standard deviation (filter scale)
    norm : bool
        Option to return normalised data. Default (norm=True)

    Returns
    -------
    float Gaussian filtered data point

    """

    x = check_float(x)
    sigma = check_float(sigma)

    val = np.exp(-0.5 * (x / sigma) ** 2)

    if norm:
        return val / (np.sqrt(2 * np.pi) * sigma)

    else:
        return val


def mex_hat(x, sigma):
    """Mexican hat

    This method implements a Mexican hat (or Ricker) wavelet.

    Parameters
    ----------
    x : float
        Input data point
    sigma : float
        Standard deviation (filter scale)

    Returns
    -------
    float Mexican hat filtered data point

    """

    x = check_float(x)
    sigma = check_float(sigma)

    xs = (x / sigma) ** 2
    val = 2 * (3 * sigma) ** -0.5 * np.pi ** -0.25

    return val * (1 - xs) * np.exp(-0.5 * xs)


def mex_hat_dir(x, y, sigma):
    """Directional Mexican hat

    This method implements a directional Mexican hat (or Ricker) wavelet.

    Parameters
    ----------
    x : float
        Input data point for Gaussian
    y : float
        Input data point for Mexican hat
    sigma : float
        Standard deviation (filter scale)

    Returns
    -------
    float directional Mexican hat filtered data point

    """

    return -0.5 * (x / sigma) ** 2 * mex_hat(y, sigma)



# -*- coding: utf-8 -*-

"""POSITIVITY

This module contains a function that retains only positive coefficients in
an array

:Author: Samuel Farrens <samuel.farrens@gmail.com>

:Version: 1.0

:Date: 26/07/2017

"""

import numpy as np


def positive(data):
    """Positivity operator

    This method preserves only the positive coefficients of the input data, all
    negative coefficients are set to zero

    Parameters
    ----------
    data : np.ndarray, list or tuple
        Input data array

    Returns
    -------
    np.ndarray array with only positive coefficients

    """

    def pos_recursive(data):

        data = np.array(data)

        if not data.dtype == 'O':

            result = list(data * (data > 0))

        else:

            result = [pos_recursive(x) for x in data]

        return result

    return np.array(pos_recursive(data))



# -*- coding: utf-8 -*-

"""WAVELET MODULE

This module contains methods for performing wavelet transformations using iSAP

:Author: Samuel Farrens <samuel.farrens@gmail.com>

:Version: 1.2

:Date: 20/10/2017

"""

from __future__ import division
from builtins import zip
import numpy as np
from os import remove
from subprocess import check_call
from datetime import datetime
from astropy.io import fits
from sf_tools.image.convolve import convolve
from sf_tools.base.np_adjust import rotate_stack


def call_mr_transform(data, opt=None, path='./', remove_files=True):
    """Call mr_transform

    This method calls the iSAP module mr_transform

    Parameters
    ----------
    data : np.ndarray
        Input data, 2D array
    opt : list, optional
        List of additonal mr_transform options
    path : str, optional
        Path for output files (default is './')
    remove_files : bool, optional
        Option to remove output files (default is 'True')

    Returns
    -------
    np.ndarray results of transform

    """

    # Create a unique string using the current date and time.
    unique_string = datetime.now().strftime('%Y.%m.%d_%H.%M.%S')

    # Set the ouput file names.
    file_name = path + 'mr_temp_' + unique_string
    file_fits = file_name + '.fits'
    file_mr = file_name + '.mr'

    # Write the input data to a fits file.
    fits.writeto(file_fits, data)

    # Call mr_transform.
    if isinstance(opt, type(None)):
        check_call(['mr_transform', file_fits, file_mr])
    else:
        check_call(['mr_transform'] + opt + [file_fits, file_mr])

    # Retrieve wavelet transformed data.
    result = fits.getdata(file_mr)

    # Return the mr_transform results (and the output file names).
    if remove_files:
        remove(file_fits)
        remove(file_mr)
        return result
    else:
        return result, file_mr


def get_mr_filters(data_shape, opt=None, coarse=False):
    """Get mr_transform filters

    This method obtains wavelet filters by calling mr_transform

    Parameters
    ----------
    data_shape : tuple
        2D data shape
    opt : list, optional
        List of additonal mr_transform options
    coarse : bool, optional
        Option to keep coarse scale (default is 'False')

    Returns
    -------
    np.ndarray 3D array of wavelet filters

    """

    # Adjust the shape of the input data.
    data_shape = np.array(data_shape)
    data_shape += data_shape % 2 - 1

    # Create fake data.
    fake_data = np.zeros(data_shape)
    fake_data[list(zip(data_shape // 2))] = 1

    # Call mr_transform.
    mr_filters = call_mr_transform(fake_data, opt=opt)

    # Return filters
    if coarse:
        return mr_filters
    else:
        return mr_filters[:-1]


def filter_convolve(data, filters, filter_rot=False):
    """Filter convolve

    This method convolves the input image with the wavelet filters

    Parameters
    ----------
    data : np.ndarray
        Input data, 2D array
    filters : np.ndarray
        Wavelet filters, 3D array
    filter_rot : bool, optional
        Option to rotate wavelet filters (default is 'False')

    Returns
    -------
    np.ndarray convolved data

    """

    if filter_rot:
        return np.sum((convolve(coef, f) for coef, f in
                      zip(data, rotate_stack(filters))), axis=0)

    else:
        return np.array([convolve(data, f) for f in filters])


def filter_convolve_stack(data, filters, filter_rot=False):
    """Filter convolve

    This method convolves the a stack of input images with the wavelet filters

    Parameters
    ----------
    data : np.ndarray
        Input data, 3D array
    filters : np.ndarray
        Wavelet filters, 3D array
    filter_rot : bool, optional
        Option to rotate wavelet filters (default is 'False')

    Returns
    -------
    np.ndarray convolved data

    """

    # Return the convolved data cube.
    return np.array([filter_convolve(x, filters, filter_rot=filter_rot)
                     for x in data])



# -*- coding: utf-8 -*-

"""IMAGE STATISTICS ROUTINES

This module contains methods for the statistical analysis of images.

:Author: Samuel Farrens <samuel.farrens@gmail.com>

:Version: 1.1

:Date: 20/10/2017

"""

from __future__ import division
import numpy as np
from sf_tools.base.np_adjust import pad2d


class SAT():
    """Summed Area Table

    This class produces a Summed Area Table (SAT) for fast and efficient
    statistics on image patches.

    Parameters
    ----------
    data : np.ndarray
        Input 2D data array

    Notes
    -----
    Also know as Itegral Image (i in the class features).

    TODO
    ----
    Add equations and citations

    """

    def __init__(self, data):

        self.x = data
        self.x2 = self.x ** 2
        self.i = self.x.cumsum(axis=0).cumsum(axis=1)
        self.i2 = self.x2.cumsum(axis=0).cumsum(axis=1)
        self.i_pad = pad2d(self.i, 1)
        self.i2_pad = pad2d(self.i2, 1)

    def get_area(self, data, corners):
        """Get area

        This method calculates the area of a patch.

        Parameters
        ----------
        data : np.ndarray
            Input 2D data array
        corners : tuple
            Positions of upper left and bottom right corners.

        Returns
        -------
        float area

        """

        corners = np.array(corners)
        corners[1] += 1

        a = data[zip(corners[0])]
        b = data[corners[0, 0], corners[1, 1]]
        c = data[corners[1, 0], corners[0, 1]]
        d = data[zip(corners[1])]

        return float(a + d - b - c)

    def get_npx(self, corners):
        """Get number of pixels

        This method calculates the number of pixels in a patch.

        Parameters
        ----------
        corners : tuple
            Positions of upper left and bottom right corners.

        Returns
        -------
        int number of pixels

        """

        return np.prod(np.diff(corners, axis=0) + 1)

    def get_var(self):
        """Get variance

        This method calculates the variance and standard deviation of a patch.

        """

        self.var = (self.area2 - (self.area ** 2 / self.npx)) / self.npx
        self.std = np.sqrt(self.var)

    def set_patch(self, corners):
        """Set patch

        This method sets the corner positions of a single patch.

        Parameters
        ----------
        corners : tuple
            Positions of upper left and bottom right corners.

        """

        self.area = self.get_area(self.i_pad, corners)
        self.area2 = self.get_area(self.i2_pad, corners)
        self.npx = self.get_npx(corners)
        self.get_var()

    def set_patches(self, corners):
        """Set patches

        This method sets the corner positions for multiple patches.

        Parameters
        ----------
        corners : list
            List of the positions of upper left and bottom right corners.

        """

        self.area = np.array([self.get_area(self.i_pad, corner)
                              for corner in corners])
        self.area2 = np.array([self.get_area(self.i2_pad, corner)
                               for corner in corners])
        self.npx = np.array([self.get_npx(corner) for corner in corners])
        self.get_var()



# -*- coding: utf-8 -*-

"""PROXIMITY OPERATORS

This module contains classes of proximity operators for optimisation

:Author: Samuel Farrens <samuel.farrens@gmail.com>

:Version: 1.3

:Date: 19/07/2017

"""

from builtins import range
import numpy as np
from sf_tools.signal.noise import thresh
from sf_tools.signal.svd import svd_thresh, svd_thresh_coef
from sf_tools.signal.optimisation import ForwardBackward
from sf_tools.signal.positivity import positive
from sf_tools.base.transform import *


class Positive(object):
    """Positivity proximity operator

    This class defines the positivity proximity operator

    """

    def __init__(self):
        pass

    def op(self, data, **kwargs):
        """Operator

        This method preserves only the positive coefficients of the input data

        Parameters
        ----------
        data : np.ndarray
            Input data array

        Returns
        -------
        np.ndarray all positive elements from input data

        """

        return positive(data)


class Threshold(object):
    """Threshold proximity operator

    This class defines the threshold proximity operator

    Parameters
    ----------
    weights : np.ndarray
        Input array of weights
    thresh_type : str {'hard', 'soft'}, optional
        Threshold type (default is 'soft')

    """

    def __init__(self, weights, thresh_type='soft'):

        self.weights = weights
        self.thresh_type = thresh_type

    def op(self, data, extra_factor=1.0):
        """Operator

        This method returns the input data thresholded by the weights

        Parameters
        ----------
        data : np.ndarray
            Input data array
        extra_factor : float
            Additional multiplication factor

        Returns
        -------
        np.ndarray thresholded data

        """

        threshold = self.weights * extra_factor

        return thresh(data, threshold, self.thresh_type)


class LowRankMatrix(object):
    """Low-rank proximity operator

    This class defines the low-rank proximity operator

    Parameters
    ----------
    thresh : float
        Threshold value
    treshold_type : str {'hard', 'soft'}
        Threshold type (options are 'hard' or 'soft')
    lowr_type : str {'standard', 'ngole'}
        Low-rank implementation (options are 'standard' or 'ngole')
    operator : class
        Operator class ('ngole' only)

    """

    def __init__(self, thresh, thresh_type='soft',
                 lowr_type='standard', operator=None):

        self.thresh = thresh
        self.thresh_type = thresh_type
        self.lowr_type = lowr_type
        self.operator = operator

    def op(self, data, extra_factor=1.0):
        """Operator

        This method returns the input data after the singular values have been
        thresholded

        Parameters
        ----------
        data : np.ndarray
            Input data array
        extra_factor : float
            Additional multiplication factor

        Returns
        -------
        np.ndarray SVD thresholded data

        """

        # Update threshold with extra factor.
        threshold = self.thresh * extra_factor

        if self.lowr_type == 'standard':
            data_matrix = svd_thresh(cube2matrix(data), threshold,
                                     thresh_type=self.thresh_type)

        elif self.lowr_type == 'ngole':
            data_matrix = svd_thresh_coef(data, self.operator,
                                          threshold,
                                          thresh_type=self.thresh_type)

        new_data = matrix2cube(data_matrix, data.shape[1:])

        # Return updated data.
        return new_data


class ProximityCombo(object):
    """Proximity Combo

    This class defines a combined proximity operator

    Parameters
    ----------
    operators : list
        List of proximity operator class instances

    """

    def __init__(self, operators):

        self.operators = operators

    def op(self, data, extra_factor=1.0):
        """Operator

        This method returns the result of applying all of the proximity
        operators to the data

        Parameters
        ----------
        data : np.ndarray
            Input data array
        extra_factor : float
            Additional multiplication factor

        Returns
        -------
        np.ndarray result

        """

        res = np.empty(len(self.operators), dtype=np.ndarray)

        for i in range(len(self.operators)):
            res[i] = self.operators[i].op(data[i], extra_factor=extra_factor)

        return res


class SubIter(object):
    """Sub iteration operator

    This class defines the sub-iteration proximity operator

    Parameters
    ----------
    data_shape : tuple
        Shape of input data array
    operator : class
        Proximity operator class
    weights : np.ndarray
        Array of weights
    u_init : np.ndarray
        Initial estimate of u

    """

    def __init__(self, data_shape, operator, weights=None, u_init=None):

        self.operator = operator

        if not isinstance(weights, type(None)):
            self.weights = weights

        if isinstance(u_init, type(None)):
            self.u = np.ones(data_shape)

        self.opt = ForwardBackward(self.u, self.operator,
                                   Threshold(self.weights), auto_iterate=False,
                                   indent_level=2)

    def update_weights(self, weights):
        """Update weights

        This method updates the values of the weights

        Parameters
        ----------
        weights : np.ndarray
            Array of weights

        """

        self.weights = weights

    def update_u(self):
        """Update u

        This method updates the values of u

        """

        self.opt.iterate(100)
        self.u = self.opt.x_final

    def op(self, data):
        """Operator

        Parameters
        ----------
        data : np.ndarray
            Input data array

        Returns
        -------
        np.ndarray result

        """

        self.update_u()

        new_data = data - self.operator.adj_op(self.u)

        return new_data



# -*- coding: utf-8 -*-

"""INTEGRATION ROUTINES

This module contains methods for integration.

:Author: Samuel Farrens <samuel.farrens@gmail.com>

:Version: 1.0

:Date: 03/04/2017

"""


import numpy as np
from scipy.integrate import quad


def integrate(func, lim_low, lim_up, *args):
    """Integrate

    This method integrates a given function, which has N additional arguments,
    between the specified limits.

    Parameters
    ----------
    func : function
        Function to be integrated
    lim_low : float
        Lower limit
    lim_up : float
        Upper limit

    Returns
    -------
    Result of the definite integral

    """

    return quad(func, lim_low, lim_up, args=args)[0]


def vintegrate(func, lim_low, lim_up, *args):
    """Vectorised integration

    This method implements a vectorised version of integrate().

    Parameters
    ----------
    func : function
        Function to be integrated
    lim_low : float
        Lower limit
    lim_up : float
        Upper limit

    Returns
    -------
    Result of the definite integral

    """

    v_integ = np.vectorize(integrate)

    return v_integ(func, lim_low, lim_up, *args)



# -*- coding: utf-8 -*-

"""INTERFACE ROUTINES

This module contains submodules for error handling, logging and IO interaction.

:Author: Samuel Farrens <samuel.farrens@gmail.com>

:Version: 1.0

:Date: 06/04/2017

"""



# -*- coding: utf-8 -*-

"""IMAGE DISTORTION ROUTINES

This module contains methods for playing around with image properties.

:Author: Samuel Farrens <samuel.farrens@gmail.com>

:Version: 1.1

:Date: 20/10/2017

"""

from __future__ import division
import numpy as np
from itertools import product
from sf_tools.base.np_adjust import pad2d


def downsample(image, factor):
    """Downsample

    This method downsamples (decimates) an image.

    Parameters
    ----------
    image : np.ndarray
        Input image array
    factor : int
        Downsampling factor

    Returns
    -------
    np.ndarray downsampled image array

    Raises
    ------
    ValueError
        For invalid downsampling factor

    """

    factor = np.array(factor)

    if not np.all(factor > 0):
        raise ValueError('The downsampling factor values must be > 0.')

    if factor.size == 1:
        return image[0::factor, 0::factor]

    elif factor.size == 2:
        return image[0::factor[0], 0::factor[1]]

    else:
        raise ValueError('The downsampling factor can only contain one or ' +
                         'two values.')


def resize_even_image(image):
    """Resize even image

    This method returns an image with odd dimensions.

    Parameters
    ----------
    image : np.ndarray
        Input image array

    Returns
    -------
    np.ndarray resized image array

    """

    return image[[slice(x) for x in (np.array(image.shape) +
                  np.array(image.shape) % 2 - 1)]]


def roll_2d(data, roll_rad=(1, 1)):
    """Roll in 2D

    This method rolls an array in 2 dimensions.

    Parameters
    ----------
    data : np.ndarray
        Input 2D data array
    roll_rad : tuple
        Roll radius in each dimension

    Returns
    -------
    np.ndarray rolled array

    """

    return np.roll(np.roll(data, roll_rad[1], axis=1), roll_rad[0], axis=0)


def rot_and_roll(data):
    """Rotate and roll

    This method rotates (by 180 deg) and rolls a 2D array.

    Parameters
    ----------
    data : np.ndarray
        Input 2D data array

    Returns
    -------
    np.ndarray rotated and rolled array

    """

    return roll_2d(np.rot90(data, 2), -(np.array(data.shape) // 2))


def gen_mask(kernel_shape, image_shape):
    """Generate mask

    This method generates an image mask.

    Parameters
    ----------
    kernel_shape : tuple
        Shape of kernel
    image_shape : tuple
        Shape of image

    Returns
    -------
    np.ndarray boolean mask

    """

    kernel_shape = np.array(kernel_shape)
    image_shape = np.array(image_shape)

    shape_diff = image_shape - kernel_shape

    mask = np.ones(image_shape, dtype=bool)

    if shape_diff[0] > 0:
        mask[-shape_diff[0]:] = False
    if shape_diff[1] > 0:
        mask[:, -shape_diff[1]:] = False

    return roll_2d(mask, -(kernel_shape // 2))


def roll_sequence(data_shape):
    """Roll sequence

    This method generates the roll sequence for a 2D array.

    Parameters
    ----------
    data_shape : tuple
        Shape of data

    Returns
    -------
    list of roll radii

    """

    data_shape = np.array(data_shape)

    return list(product(*(range(data_shape[0]), range(data_shape[1]))))


def kernel_pattern(kernel_shape, mask):
    """Kernel pattern

    This method generates the kernel pattern. Rather than padding the kernel
    with zeroes to match the image size one simply extracts the series of
    repitions of the base kernel patterns.

    Parameters
    ----------
    kernel_shape : tuple
        Shape of kernel
    mask : np.ndarray
        Boolean mask

    Returns
    -------
    np.ndarray kernel pattern

    """

    kernel_shape = np.array(kernel_shape)

    kernel_buffer = 1 - np.array(kernel_shape) % 2

    n_rep_axis1 = sum(1 - mask[:, 0])
    n_rep_axis2 = sum(1 - mask[0])

    if np.any(mask[:, 0] is False):
        pos_1 = np.where(mask[:, 0] is False)[0][0] - 1 + kernel_buffer[0]

    if np.any(mask[0] is False):
        pos_2 = np.where(mask[0] is False)[0][0] - 1 + kernel_buffer[1]

    pattern = np.arange(np.prod(kernel_shape)).reshape(kernel_shape)

    for i in range(n_rep_axis1):
        pattern = np.insert(pattern, pos_1, pattern[pos_1], axis=0)

    for i in range(n_rep_axis2):
        pattern = np.insert(pattern, pos_2, pattern[:, pos_2], axis=1)

    return pattern.reshape(pattern.size)


def rearrange_kernel(kernel, data_shape=None):
    """Rearrange kernel

    This method rearanges the input kernel elements for vector multiplication.
    The input kernel is padded with zeroes to match the image size.

    Parameters
    ----------
    kernel : np.ndarray
        Input kernel array
    data_shape : tuple
        Shape of the data

    Returns
    -------
    np.ndarray rearanged matrix of kernel elements

    """

    # Define kernel shape.
    kernel_shape = np.array(kernel.shape)

    # Set data shape if not provided.
    if isinstance(data_shape, type(None)):
        data_shape = kernel_shape
    else:
        data_shape = np.array(data_shape)

    # Set the length of the output matrix rows.
    vec_length = np.prod(data_shape)

    # Find the diffrence between the shape of the data and the kernel.
    shape_diff = data_shape - kernel_shape

    if np.any(shape_diff < 0):
        raise ValueError('Kernel shape must be less than or equal to the '
                         'data shape')

    # Set the kernel radius.
    kernel_rad = kernel_shape // 2

    # Rotate, pad and roll the input kernel.
    kernel_rot = np.pad(np.rot90(kernel, 2), ((0, shape_diff[0]),
                        (0, shape_diff[1])), 'constant')
    kernel_rot = np.roll(np.roll(kernel_rot, -kernel_rad[1], axis=1),
                         -kernel_rad[0], axis=0)

    return np.array([np.roll(np.roll(kernel_rot, i, axis=0), j,
                    axis=1).reshape(vec_length) for i in range(data_shape[0])
                    for j in range(data_shape[1])])



# -*- coding: utf-8 -*-

"""SHAPE ESTIMATION ROUTINES

This module contains methods and classes for estimating galaxy shapes.

:Author: Samuel Farrens <samuel.farrens@gmail.com>

:Version: 1.4

:Date: 20/10/2017

Notes
-----
Some of the methods in this module are based on work by Fred Ngole.

"""

from __future__ import division
import numpy as np


def ellipticity_atoms(data, offset=0):
    r"""Calculate ellipticity

    This method calculates the ellipticity of an image from its shape
    projection components.

    Parameters
    ----------
    data : np.ndarray
        Input data array, the image to be analysed
    offset : int, optional
        Shape projection offset (default is '0')

    Returns
    -------
    np.ndarray of the image ellipticity components

    See Also
    --------
    shape_project : shape projection matrix

    Notes
    -----
    This technique was developed by Fred Ngole and implements the following
    equations:

        - Equations C.1 and C.2 from [NS2016]_ appendix:

        .. math::

            e_1(\mathbf{X}_i) = \frac{<\mathbf{X}_i, \mathbf{U}_4>
                                      <\mathbf{X}_i, \mathbf{U}_2> -
                                      <\mathbf{X}_i, \mathbf{U}_0>^2 +
                                      <\mathbf{X}_i, \mathbf{U}_1>^2}
                                      {<\mathbf{X}_i, \mathbf{U}_3>
                                      <\mathbf{X}_i, \mathbf{U}_2> -
                                      <\mathbf{X}_i, \mathbf{U}_0>^2 -
                                      <\mathbf{X}_i, \mathbf{U}_1>^2
                                      }

            e_2(\mathbf{X}_i) = \frac{2\left(<\mathbf{X}_i, \mathbf{U}_5>
                                      <\mathbf{X}_i, \mathbf{U}_2> -
                                      <\mathbf{X}_i, \mathbf{U}_0>
                                      <\mathbf{X}_i, \mathbf{U}_1>\right)}
                                      {<\mathbf{X}_i, \mathbf{U}_3>
                                      <\mathbf{X}_i, \mathbf{U}_2> -
                                      <\mathbf{X}_i, \mathbf{U}_0>^2 -
                                      <\mathbf{X}_i, \mathbf{U}_1>^2
                                      }

    Examples
    --------
    >>> from image.shape import ellipticity_atoms
    >>> import numpy as np
    >>> a = np.zeros((5, 5))
    >>> a[2, 1:4] += 1
    >>> ellipticity_atoms(a)
    array([-1.,  0.])

    >>> b = np.zeros((5, 5))
    >>> b[1:4, 2] += 1
    >>> ellipticity_atoms(b)
    array([ 1.,  0.])

    """

    XU = [np.sum(data * U) for U in shape_project(data.shape, offset)]

    divisor = XU[3] * XU[2] - XU[0] ** 2 - XU[1] ** 2
    e1 = (XU[4] * XU[2] - XU[0] ** 2 + XU[1] ** 2) / divisor
    e2 = 2 * (XU[5] * XU[2] - XU[0] * XU[1]) / divisor

    return np.array([e1, e2])


def shape_project(shape, offset=0, return_norm=False):
    r"""Shape projection matrix

    This method generates a shape projection matrix for a given image.

    Parameters
    ----------
    shape : list, tuple or np.ndarray
        List of image dimensions
    offset : int, optional
        Shape projection offset (default is '0')
    return_norm : bool, optional
        Option to return l2 normalised shape projection components
        (default is 'False')

    Returns
    -------
    np.ndarray of shape projection components

    See Also
    --------
    ellipticity_atoms : calculate ellipticity

    Notes
    -----
    This technique was developed by Fred Ngole and implements the following
    equations:

        - Equations from [NS2016]_ appendix:

        .. math::

            U_1 &= (k)_{1 \leq k \leq N_l, 1 \leq l \leq N_c} \\
            U_2 &= (l)_{1 \leq k \leq N_l, 1 \leq l \leq N_c} \\
            U_3 &= (1)_{1 \leq k \leq N_l, 1 \leq l \leq N_c} \\
            U_4 &= (k^2 + l^2)_{1 \leq k \leq N_l, 1 \leq l \leq N_c} \\
            U_5 &= (k^2 - l^2)_{1 \leq k \leq N_l, 1 \leq l \leq N_c} \\
            U_6 &= (kl)_{1 \leq k \leq N_l, 1 \leq l \leq N_c}

    Examples
    --------
    >>> from image.shape import shape_project
    >>> shape_project([3, 3])
    array([[[ 0.,  0.,  0.],
            [ 1.,  1.,  1.],
            [ 2.,  2.,  2.]],
    <BLANKLINE>
           [[ 0.,  1.,  2.],
            [ 0.,  1.,  2.],
            [ 0.,  1.,  2.]],
    <BLANKLINE>
           [[ 1.,  1.,  1.],
            [ 1.,  1.,  1.],
            [ 1.,  1.,  1.]],
    <BLANKLINE>
           [[ 0.,  1.,  4.],
            [ 1.,  2.,  5.],
            [ 4.,  5.,  8.]],
    <BLANKLINE>
           [[ 0., -1., -4.],
            [ 1.,  0., -3.],
            [ 4.,  3.,  0.]],
    <BLANKLINE>
           [[ 0.,  0.,  0.],
            [ 0.,  1.,  2.],
            [ 0.,  2.,  4.]]])

    """

    U = []
    U.append(np.outer(np.arange(shape[0]) + offset, np.ones(shape[1])))
    U.append(np.outer(np.ones(shape[0]), np.arange(shape[1]) + offset))
    U.append(np.ones(shape))
    U.append(U[0] ** 2 + U[1] ** 2)
    U.append(U[0] ** 2 - U[1] ** 2)
    U.append(U[0] * U[1])

    if return_norm:
        np.array([np.linalg.norm(x, 2) for x in U])
    else:
        return np.array(U)


class Ellipticity():
    """ Image ellipticity class

    This class calculates image ellipticities from quadrupole moments.

    Examples
    --------
    >>> from image.shape import Ellipticity
    >>> import numpy as np
    >>> a = np.zeros((5, 5))
    >>> a[2, 1:4] += 1
    >>> Ellipticity(a).e
    array([-1.,  0.])

    >>> b = np.zeros((5, 5))
    >>> b[1:4, 2] += 1
    >>> Ellipticity(b).e
    array([ 1.,  0.])

    """

    ##
    #  Method that initialises the class.
    #
    def __init__(self, data, sigma=1000, centroid=None, moments=None):
        """Class initialiser

        Parameters
        ----------
        data : np.ndarray
            Input data array, the image to be analysed
        sigma : int, optional
            Estimation error (default is '1000')
        centroid : np.ndarray, optional
            Centroid positions [x, y] of the input image (defualt is 'None')
        moments : np.ndarray, optional
            Quadrupole moments [[q00, q01], [q10, q11]] of the input image
            (defualt is 'None')

        """

        self.data = data
        self.sigma = sigma
        self.ranges = np.array([np.arange(i) for i in data.shape])

        if not isinstance(moments, type(None)):
            self.moments = np.array(moments).reshape(2, 2)
            self.get_ellipse()
        elif isinstance(centroid, type(None)):
            self.get_centroid()
        else:
            self.centroid = centroid
            self.update_weights()
            self.get_moments()

    def update_xy(self):
        """Update the x and y values

        This method updates the values of x and y using the current centroid.

        """

        self.x = np.outer(self.ranges[0] - self.centroid[0],
                          np.ones(self.data.shape[1]))
        self.y = np.outer(np.ones(self.data.shape[0]),
                          self.ranges[1] - self.centroid[1])

    def update_weights(self):
        """Update the weights

        This method updates the value of the weights using the current values
        of x and y.

        Notes
        -----
        This method implements the following equations:

            - The exponential part of equation 1 from [BM2007]_ to calculate
              the weights:

            .. math::

                w(x,y) = e^{-\\frac{\\left((x-x_c)^2+(y-y_c)^2\\right)}
                    {2\\sigma^2}}

        """

        self.update_xy()
        self.weights = np.exp(-(self.x ** 2 + self.y ** 2) /
                              (2 * self.sigma ** 2))

    def update_centroid(self):
        """Update the centroid

        This method updates the centroid value using the current weights.

        Notes
        -----
        This method implements the following equations:

            - Equation 2a, 2b and 2c from [BM2007]_ to calculate the position
              moments:

            .. math::

                S_w = \sum_{x,y} I(x,y)w(x,y)

                S_x = \sum_{x,y} xI(x,y)w(x,y)

                S_y = \sum_{x,y} yI(x,y)w(x,y)

            - Equation 3 from [BM2007]_ to calculate the centroid:

            .. math::

                X_c = S_x/S_w,\\
                Y_c = S_y/S_w

        """

        # Calculate the position moments.
        iw = np.array([np.sum(self.data * self.weights, axis=i)
                       for i in (1, 0)])
        sw = np.sum(iw, axis=1)
        sxy = np.sum(iw * self.ranges, axis=1)

        # Update the centroid value.
        self.centroid = sxy / sw

    def get_centroid(self, n_iter=10):
        """Calculate centroid

        This method iteratively calculates the centroid of the image.

        Parameters
        ----------
        n_iter : int, optional
            Number of iterations (deafult is '10')

        """

        # Set initial value for the weights.
        self.weights = np.ones(self.data.shape)

        # Iteratively calculate the centroid.
        for i in range(n_iter):

            # Update the centroid value.
            self.update_centroid()

            # Update the weights.
            self.update_weights()

        # Calculate the quadrupole moments.
        self.get_moments()

    def get_moments(self):
        """ Calculate the quadrupole moments

        This method calculates the quadrupole moments.

        Notes
        -----
        This method implements the following equations:

            - Equation 10 from [C2013]_ to calculate the moments:

            .. math::

                Q_{ij}=\\frac{\\int\\int\\Phi(x_i,x_j) w(x_i,x_j)
                    (x_i-\\bar{x_i})(x_j-\\bar{x_j}) dx_idx_j}
                    {\\int\\int\\Phi(x_i,x_j)w(x_i,x_j)dx_idx_j}

        """

        # Calculate moments.
        q = np.array([np.sum(self.data * self.weights * xi * xj) for xi in
                      (self.x, self.y) for xj in (self.x, self.y)])

        self.moments = (q / np.sum(self.data * self.weights)).reshape(2, 2)

        # Calculate the ellipticities.
        self.get_ellipse()

    def get_ellipse(self):
        """Calculate the ellipticities

        This method cacluates ellipticities from quadrupole moments.

        Notes
        -----
        This method implements the following equations:

            - Equation 11 from [C2013]_ to calculate the size:

            .. math:: R^2 = Q_{00} + Q_{11}

            - Equation 12 from [C2013]_ to calculate the ellipticities:

            .. math::

               \\varepsilon = [\\varepsilon_1,\\varepsilon_2] =
                  \\left[\\frac{Q_{00}-Q_{11}}{R^2},
                  \\frac{Q_{01}+Q_{10}}{R^2}\\right]

        """

        # Calculate the size.
        self.r2 = self.moments[0, 0] + self.moments[1, 1]

        # Calculate the ellipticities.
        self.e = np.array([(self.moments[0, 0] - self.moments[1, 1]) / self.r2,
                          (self.moments[0, 1] + self.moments[1, 0]) /
                          self.r2])



# -*- coding: utf-8 -*-

"""IMAGE ROUTINES

This module contains submodules for image analysis.

:Author: Samuel Farrens <samuel.farrens@gmail.com>

:Version: 1.0

:Date: 06/04/2017

References
----------

.. [C2013] Cropper et al., Defining a Weak Lensing Experiment in Space, 2013,
    MNRAS, 431, 3103C. [https://arxiv.org/abs/1210.7691]

.. [BM2007] Baker and Moallem, Iteratively weighted centroiding for
    Shack-Hartman wave-front sensors, 2007n, Optics Express, 15, 8, 5147.
    [https://www.osapublishing.org/oe/abstract.cfm?uri=oe-15-8-5147]

.. [NS2016] Ngolé and Starck, PSFs field learning based on Optimal Transport
    distances, 2016, SIAM

"""



# -*- coding: utf-8 -*-

"""BASE ROUTINES

This module contains submodules for basic operations such as type
transformations and adjustments to the default output of Numpy functions.

:Author: Samuel Farrens <samuel.farrens@gmail.com>

:Version: 1.0

:Date: 06/04/2017

"""



# -*- coding: utf-8 -*-

"""LOGGING ROUTINES

This module contains methods for handing logging.

:Author: Samuel Farrens <samuel.farrens@gmail.com>

:Version: 1.2

:Date: 20/10/2017

"""

from __future__ import print_function
import sys
import logging


def set_up_log(filename, verbose=True):
    """Set up log

    This method sets up a basic log.

    Parameters
    ----------
    filename : str
        Log file name

    Returns
    -------
    logging.Logger instance

    """

    # Add file extension.
    filename += '.log'

    if verbose:
        print('Preparing log file:', filename)

    # Capture warnings.
    logging.captureWarnings(True)

    # Set output format.
    formatter = logging.Formatter(fmt='%(asctime)s %(message)s',
                                  datefmt='%d/%m/%Y %H:%M:%S')

    # Create file handler.
    fh = logging.FileHandler(filename=filename, mode='w')
    fh.setLevel(logging.DEBUG)
    fh.setFormatter(formatter)

    # Create log.
    log = logging.getLogger('log')
    log.setLevel(logging.DEBUG)
    log.addHandler(fh)

    # Send opening message.
    log.info('The log file has been set-up.')

    return log


def close_log(log, verbose=True):
    """Close log

    This method closes and active logging.Logger instance.

    Parameters
    ----------
    log : logging.Logger
        Logging instance

    """

    if verbose:
        print('Closing log')

    # Send closing message.
    log.info('The log file has been closed.')

    # Remove all handlers from log.
    [log.removeHandler(handler) for handler in log.handlers]



# -*- coding: utf-8 -*-

"""GRADIENT CLASSES

This module contains classses for defining algorithm gradients.
Based on work by Yinghao Ge and Fred Ngole.

:Author: Samuel Farrens <samuel.farrens@gmail.com>

:Version: 1.3

:Date: 19/07/2017

"""


class GradBasic(object):
    """Basic gradient class

    This class defines the basic methods that will be inherited by specific
    gradient classes

    """

    def MtMX(self, x):
        """M^T M X

        This method calculates the action of the transpose of the matrix M on
        the action of the matrix M on the data X

        Parameters
        ----------
        x : np.ndarray
            Input data array

        Returns
        -------
        np.ndarray result

        Notes
        -----
        Calculates  M^T (MX)

        """

        return self.MtX(self.MX(x))

    def get_grad(self, x):
        """Get the gradient step

        This method calculates the gradient step from the input data

        Parameters
        ----------
        x : np.ndarray
            Input data array

        Returns
        -------
        np.ndarray gradient value

        Notes
        -----

        Calculates M^T (MX - Y)

        """

        self.grad = self.MtX(self.MX(x) - self.y)



# -*- coding: utf-8 -*-

"""MATHEMATICS ROUTINES

This module contains submodules for mathematical applications.

:Author: Samuel Farrens <samuel.farrens@gmail.com>

:Version: 1.0

:Date: 06/04/2017

"""



# -*- coding: utf-8 -*-

"""IMAGE STAMP SELECTION ROUTINES

This module contains methods for selecting stamps or patches from
images.

:Author: Samuel Farrens <samuel.farrens@gmail.com>

:Version: 1.0

:Date: 20/10/2017

"""

from __future__ import division
from builtins import zip
import numpy as np
from itertools import product
from sf_tools.base.np_adjust import pad2d


def patch_centres(data_shape, layout):
    """Image centres

    This method inds the centres of the patches in a 2D map.

    Parameters
    ----------
    data_shape : tuple
        Shape of the 2D map
    layout : tuple
        Layout of the patches

    Returns
    -------
    np.ndarray array of patch centres

    """

    data_size = np.array(data_size)
    layout = np.array(layout)

    if data_size.size != 2:
        raise ValueError('The the data shape must be of size 2.')
    if layout.size != 2:
        raise ValueError('The the layout must be of size 2.')

    ranges = np.array(list(product(*np.array([np.arange(x) for x in layout]))))
    patch_shape = data_shape // layout
    patch_centre = patch_shape // 2

    return patch_centre + patch_size * ranges


def postage_stamp(data, pos, pixel_rad):
    """Postage stamp

    This metho selects a postage stamp of a given size from a 2D-array.

    Parameters
    ----------
    data : np.ndarray
        Input data array
    pos : tuple
        Position of postage stamp centre
    pixel_rad : tuple
        Pixel radius in each dimension (i.e. size of the stamp from centre)

    Returns
    -------
    np.ndarray rearanged matrix of kernel elements

    Notes
    -----
    The image edges are padded with zeros.

    """

    # Check that the input array has two dimensions.
    data = np.array(data)
    if data.ndim != 2:
        raise ValueError('The input array must be 2D.')

    # If the pixel radius size is one repat the value for the 2nd dimension.
    pixel_rad = np.array(pixel_rad)
    if pixel_rad.size == 1:
        pixel_rad = np.repeat(pixel_rad, 2)
    if pixel_rad.size not in (1, 2):
        raise ValueError('The pixel radius must have a size of 1 or 2.')

    # Check that the array position has a size of two.
    pos = np.array(pos) % np.array(data.shape)
    if pos.size != 2:
        raise ValueError('The array position must have a size of 2.')

    # Check if the pixel radius is within the bounds of the input array.
    if (np.any(pixel_rad < 1) or np.any(np.array(data.shape) // 2 -
                                        pixel_rad < 0)):
        raise ValueError('The pixel radius values must have a value of at '
                         'least 1 and at most half the size of the input '
                         'array. Array size: ' + str(data.shape))

    # Return postage stamp.
    return pad2d(data, pixel_rad)[[slice(a, a + 2 * b + 1) for a, b in
                                   zip(pos, pixel_rad)]]


def pixel_pos(array_shape):
    """Pixel positions

    This method returns all of the pixel positions from a 2D-array.

    Parameters
    ----------
    array_shape : tuple
        Shape of array

    Returns
    -------
    list of pixel positions

    """

    ranges = np.array([np.arange(x) for x in np.array(array_shape)])

    return list(product(*ranges))


class FetchStamps(object):
    """Fetch postage stamps

    This class returns a stack of postage stamps from a given 2D image array.

    Parameters
    ----------
    data : np.ndarray
        Input 2D data array
    pixel_rad : tuple
        Pixel radius in each dimension
    all : boolean, optional
        Option to select all pixels. Default (all=False)

    """

    def __init__(self, data, pixel_rad, all=False):
        self.data = np.array(data)
        self.shape = np.array(self.data.shape)
        self.pixel_rad = np.array(pixel_rad)
        self._check_inputs()
        self._pad_data()
        if all:
            self.n_pixels()

    def _check_inputs(self):
        """Check inputs

        This method checks the class variable values.

        Raises
        ------
        ValueError
            For invalid array dimensions or pixel radius values

        """

        if self.data.ndim != 2:
            raise ValueError('The input array must be 2D.')
        if self.pixel_rad.size == 1:
            self.pixel_rad = np.repeat(self.pixel_rad, 2)
        elif self.pixel_rad.size not in (1, 2):
            raise ValueError('The pixel radius must have a size of 1 or 2.')

    def _pad_data(self):
        """Pad data

        This method pads the input array with zeros.

        """

        self.pad_data = pad2d(self.data, self.pixel_rad)

    def _adjust_pixels(self):
        """Adjust pixels

        This method adjusts the pixel positions according to the pixel radius.

        """

        self.pixels = self.pixels % self.shape + self.pixel_rad

    def get_pixels(self, pixels):
        """Get pixels

        This method gets the desired pixel positions.

        Parameters
        ----------
        pixels : list or np.ndarray
            List of pixel positions

        Raises
        ------
        ValueError
            For invalid number of dimensions for pixel position array

        """

        self.pixels = np.array(pixels)
        if not 1 <= self.pixels.ndim <= 2:
            raise ValueError('Invalid number of dimensions for pixels')
        elif self.pixels.ndim == 2 and self.pixels.shape[1] != 2:
            raise ValueError('The second dimension of pixels must have size 2')
        self._adjust_pixels()

    def n_pixels(self, n_pixels=None, random=False):
        """Number of pixels

        This method selects a specified number of pixel positions.

        Parameters
        ----------
        n_pixels : int, optional
            Number of pixels to keep. Default (n_pixels=None)
        random : bool, optional
            Option to select random pixel position

        """

        self.pixels = pixel_pos(self.shape)
        if random:
            np.random.shuffle(self.pixels)
        self.pixels = self.pixels[:n_pixels]
        self._adjust_pixels()

    def _stamp(self, pos, func=None, *args):
        """Stamp

        This method retrieves a postage stamp from the padded input data at a
        given position.

        Parameters
        ----------
        pos : tuple
            Pixel position in 2D padded array.
        func : function, optional
            Optional function to be applied to postage stamp.

        Returns
        -------
        np.ndarray postage stamp array or result of func

        Raises
        ------
        ValueError
            For for invalid size of pixel position.

        """

        pos = np.array(pos)
        if pos.size != 2:
            raise ValueError('The pixel position must have a size of 2.')
        stamp = self.pad_data[[slice(a - b, a + b + 1) for a, b in
                              zip(pos, self.pixel_rad)]]
        if isinstance(func, type(None)):
            return stamp
        else:
            return func(stamp, *args)

    def scan(self, func=None, *args, **kwargs):
        """Scan stamps

        This method scans the 2D padded input array and retrieves the postage
        stamps at all the desired pixel positions.

        Parameters
        ----------
        func : function, optional
            Optional function to be applied to postage stamps.

        Returns
        -------
        np.ndarray postage stamp arrays or results of func

        """

        if 'arg_type' in kwargs and kwargs['arg_type'] == 'list':
            return np.array([self._stamp(pos, func, arg) for pos, arg in
                             zip(self.pixels, *args)])

        else:
            return np.array([self._stamp(pos, func, *args) for pos in
                             self.pixels])



"""MATRIX ROUTINES

This module contains methods for matrix operations.

:Author: Samuel Farrens <samuel.farrens@gmail.com>

:Version: 1.2

:Date: 20/10/2017

"""

from __future__ import division
from builtins import range, zip
import numpy as np
from itertools import product


def gram_schmidt(matrix, return_opt='orthonormal'):
    r"""Gram-Schmit

    This method orthonormalizes the row vectors of the input matrix.

    Parameters
    ----------
    matrix : np.ndarray
        Input matrix array
    return_opt : str {orthonormal, orthogonal, both}
        Option to return u, e or both.

    Returns
    -------
    Lists of orthogonal vectors, u, and/or orthonormal vectors, e

    TODO
    ----
    Add citation and equation

    """

    u = []
    e = []

    for vector in matrix:

        if len(u) == 0:
            u_now = vector
        else:
            u_now = vector - sum([project(u_i, vector) for u_i in u])

        u.append(u_now)
        e.append(u_now / np.linalg.norm(u_now, 2))

    u = np.array(u)
    e = np.array(e)

    if return_opt == 'orthonormal':
        return e
    elif return_opt == 'orthogonal':
        return u
    else:
        return u, e


def nuclear_norm(data):
    """Nuclear norm

    This method computes the nuclear (or trace) norm of the input data.

    Parameters
    ----------
    data : np.ndarray
        Input data array

    Returns
    -------
    float nuclear norm value

    TODO
    ----
    Add equation

    """

    # Get SVD of the data.
    u, s, v = np.linalg.svd(data)

    # Return nuclear norm.
    return np.sum(s)


def project(u, v):
    """Project vector

    This method projects vector v onto vector u.

    Parameters
    ----------
    u : np.ndarray
        Input vector
    v : np.ndarray
        Input vector

    Returns
    -------
    np.ndarray projection

    TODO
    ----
    Add equation

    """

    return np.inner(v, u) / np.inner(u, u) * u


def rot_matrix(angle):
    """Rotation matrix

    This method produces a 2x2 rotation matrix for the given input angle.

    Parameters
    ----------
    angle : float
        Rotation angle

    Returns
    -------
    np.ndarray 2x2 rotation matrix

    """

    return np.around(np.array([[np.cos(angle), -np.sin(angle)],
                     [np.sin(angle), np.cos(angle)]], dtype='float'), 10)


def rotate(matrix, angle):
    """Rotate

    This method rotates an input matrix about the input angle.

    Parameters
    ----------
    matrix : np.ndarray
        Input matrix array
    angle : float
        Rotation angle

    Returns
    -------
    np.ndarray rotated matrix

    Raises
    ------
    ValueError
        For invalid matrix shape

    """

    shape = np.array(matrix.shape)

    if shape[0] != shape[1]:
        raise ValueError('Input matrix must be square.')

    shift = (np.array(shape) - 1) // 2

    index = np.array(list(product(*np.array([np.arange(val) for val in
                     shape])))) - shift

    new_index = np.array(np.dot(index, rot_matrix(angle)), dtype='int') + shift
    new_index[new_index >= shape[0]] -= shape[0]

    return matrix[list(zip(new_index.T))].reshape(shape.T)


class PowerMethod(object):
    """Power method class

    This method performs implements power method to calculate the spectral
    radius of the input data

    Parameters
    ----------
    operator : class
        Operator class instance
    data_shape : tuple
        Shape of the data array
    auto_run : bool
        Option to automatically calcualte the spectral radius upon
        initialisation

    """

    def __init__(self, operator, data_shape, auto_run=True):

        self.op = operator
        self.data_shape = data_shape
        if auto_run:
            self.get_spec_rad()

    def set_initial_x(self):
        """Set initial value of x

        This method sets the initial value of x to an arrray of random values

        """

        return np.random.random(self.data_shape)

    def get_spec_rad(self, tolerance=1e-6, max_iter=10):
        """Get spectral radius

        This method calculates the spectral radius

        Parameters
        ----------
        tolerance : float, optional
            Tolerance threshold for convergence (default is "1e-6")
        max_iter : int, optional
            Maximum number of iterations

        """

        # Set (or reset) values of x.
        x_old = self.set_initial_x()

        # Iterate until the L2 norm of x converges.
        for i in range(max_iter):

            x_new = self.op(x_old) / np.linalg.norm(x_old)

            if(np.abs(np.linalg.norm(x_new) - np.linalg.norm(x_old)) <
               tolerance):
                print(' - Power Method converged after %d iterations!' %
                      (i + 1))
                break

            elif i == max_iter - 1:
                print(' - Power Method did not converge after %d '
                      'iterations!' % max_iter)

            np.copyto(x_old, x_new)

        self.spec_rad = np.linalg.norm(x_new)
        self.inv_spec_rad = 1.0 / self.spec_rad



# -*- coding: utf-8 -*-

"""SIGNAL PROCESSING ROUTINES

This module contains submodules for signal processing.

:Author: Samuel Farrens <samuel.farrens@gmail.com>

:Version: 1.0

:Date: 06/04/2017

References
----------

.. [Con2013] Condat, A Primal-Dual Splitting Method for Convex Optimization
    Involving Lipschitzian, Proximable and Linear Composite Terms, 2013,
    Journal of Optimization Theory and Applications, 158, 2, 460.
    [https://link.springer.com/article/10.1007/s10957-012-0245-9]

.. [B2011] Bauschke et al., Fixed-Point Algorithms for Inverse Problems in
    Science and Engineering, 2011, Chapter 10.
    [http://rentals.springer.com/product/9781441995698]

.. [R2012] Raguet et al., Generalized Forward-Backward Splitting, 2012, SIAM,
    v6. [https://arxiv.org/abs/1108.4404]

.. [CWB2007] Candes, Wakin and Boyd, Enhancing Sparsity by Reweighting l1
    Minimization, 2007, Journal of Fourier Analysis and Applications,
    14(5):877-905. [https://arxiv.org/abs/0711.1612]

"""



# -*- coding: utf-8 -*-

"""DATA TRANSFORM ROUTINES

This module contains methods for transforming data.

:Author: Samuel Farrens <samuel.farrens@gmail.com>

:Version: 1.2

:Date: 20/10/2017

"""

from __future__ import division
from builtins import range
import numpy as np
from scipy.ndimage import gaussian_filter
from itertools import islice, product


def cube2map(data_cube, layout):
    """Cube to Map

    This method transforms the input data from a 3D cube to a 2D map with a
    specified layout

    Parameters
    ----------
    data_cube : np.ndarray
        Input data cube, 3D array of 2D images
    Layout : tuple
        2D layout of 2D images

    Returns
    -------
    np.ndarray 2D map

    Raises
    ------
    ValueError
        For invalid layout

    """

    if data_cube.shape[0] != np.prod(layout):
        raise ValueError('The desired layout must match the number of input '
                         'data layers.')

    return np.vstack([np.hstack(data_cube[slice(layout[1] * i, layout[1] *
                      (i + 1))]) for i in range(layout[0])])


def map2cube(data_map, layout):
    """Map to cube

    This method transforms the input data from a 2D map with given layout to
    a 3D cube

    Parameters
    ----------
    data_map : np.ndarray
        Input data map, 2D array
    layout : tuple
        2D layout of 2D images

    Returns
    -------
    np.ndarray 3D cube

    Raises
    ------
    ValueError
        For invalid layout

    """

    if np.all(np.array(data_map.shape) % np.array(layout)) != 0:
        raise ValueError('The desired layout must be a multiple of the number '
                         'pixels in the data map.')

    d_shape = np.array(data_map.shape) // np.array(layout)

    return np.array([data_map[(slice(i * d_shape[0], (i + 1) * d_shape[0]),
                    slice(j * d_shape[1], (j + 1) * d_shape[1]))] for i in
                    range(layout[0]) for j in range(layout[1])])


def map2matrix(data_map, layout):
    """Map to Matrix

    This method transforms a 2D map to a 2D matrix

    Parameters
    ----------
    data_map : np.ndarray
        Input data map, 2D array
    layout : tuple
        2D layout of 2D images

    Returns
    -------
    np.ndarray 2D matrix

    Raises
    ------
    ValueError
        For invalid layout

    """

    layout = np.array(layout)

    # Select n objects
    n_obj = np.prod(layout)

    # Get the shape of the galaxy images
    gal_shape = (np.array(data_map.shape) // layout)[0]

    # Stack objects from map
    data_matrix = []

    for i in range(n_obj):
        lower = (gal_shape * (i // layout[1]),
                 gal_shape * (i % layout[1]))
        upper = (gal_shape * (i // layout[1] + 1),
                 gal_shape * (i % layout[1] + 1))
        data_matrix.append((data_map[lower[0]:upper[0],
                            lower[1]:upper[1]]).reshape(gal_shape ** 2))

    return np.array(data_matrix).T


def matrix2map(data_matrix, map_shape):
    """Matrix to Map

    This method transforms a 2D matrix to a 2D map

    Parameters
    ----------
    data_matrix : np.ndarray
        Input data matrix, 2D array
    map_shape : tuple
        2D shape of the output map

    Returns
    -------
    np.ndarray 2D map

    Raises
    ------
    ValueError
        For invalid layout

    """

    map_shape = np.array(map_shape)

    # Get the shape and layout of the galaxy images
    gal_shape = np.sqrt(data_matrix.shape[0])
    layout = np.array(map_shape // np.repeat(gal_shape, 2), dtype='int')

    # Map objects from matrix
    data_map = np.zeros(map_shape)

    temp = data_matrix.reshape(gal_shape, gal_shape, data_matrix.shape[1])

    for i in range(data_matrix.shape[1]):
        lower = (gal_shape * (i // layout[1]),
                 gal_shape * (i % layout[1]))
        upper = (gal_shape * (i // layout[1] + 1),
                 gal_shape * (i % layout[1] + 1))
        data_map[lower[0]:upper[0], lower[1]:upper[1]] = temp[:, :, i]

    return data_map


def cube2matrix(data_cube):
    """Cube to Matrix

    This method transforms a 3D cube to a 2D matrix

    Parameters
    ----------
    data_cube : np.ndarray
        Input data cube, 3D array

    Returns
    -------
    np.ndarray 2D matrix

    """

    return data_cube.reshape([data_cube.shape[0]] +
                             [np.prod(data_cube.shape[1:])]).T


def matrix2cube(data_matrix, im_shape):
    """Matrix to Cube

    This method transforms a 2D matrix to a 3D cube

    Parameters
    ----------
    data_matrix : np.ndarray
        Input data cube, 2D array
    im_shape : tuple
        2D shape of the individual images

    Returns
    -------
    np.ndarray 3D cube

    """

    return data_matrix.T.reshape([data_matrix.shape[1]] + list(im_shape))



# -*- coding: utf-8 -*-

"""LINEAR OPERATORS

This module contains linear operator classes.

:Author: Samuel Farrens <samuel.farrens@gmail.com>

:Version: 1.3

:Date: 19/07/2017

"""

from builtins import range, zip
import numpy as np
from sf_tools.signal.wavelet import *
from sf_tools.math.matrix import rotate


class Identity(object):
    """Identity operator class

    This is a dummy class that can be used in the optimisation classes

    """

    def __init__(self):

        self.l1norm = 1.0

    def op(self, data, **kwargs):
        """Operator

        Returns the input data unchanged

        Parameters
        ----------
        data : np.ndarray
            Input data array
        **kwargs
            Arbitrary keyword arguments

        Returns
        -------
        np.ndarray input data

        """

        return data

    def adj_op(self, data):
        """Adjoint operator

        Returns the input data unchanged

        Parameters
        ----------
        data : np.ndarray
            Input data array

        Returns
        -------
        np.ndarray input data

        """

        return data


class Wavelet(object):
    """Wavelet class

    This class defines the wavelet transform operators

    Parameters
    ----------
    data : np.ndarray
        Input data array, normally an array of 2D images
    wavelet_opt: str, optional
        Additional options for `mr_transform`

    """

    def __init__(self, data, wavelet_opt=None):

        self.y = data
        self.data_shape = data.shape[-2:]
        n = data.shape[0]

        self.filters = get_mr_filters(self.data_shape, opt=wavelet_opt)
        self.l1norm = n * np.sqrt(sum((np.sum(np.abs(filter)) ** 2 for
                                       filter in self.filters)))

    def op(self, data):
        """Operator

        This method returns the input data convolved with the wavelet filters

        Parameters
        ----------
        data : np.ndarray
            Input data array, a 2D image

        Returns
        -------
        np.ndarray wavelet convolved data

        """

        return filter_convolve_stack(data, self.filters)

    def adj_op(self, data):
        """Adjoint operator

        This method returns the input data convolved with the wavelet filters
        rotated by 180 degrees

        Parameters
        ----------
        data : np.ndarray
            Input data array, a 3D of wavelet coefficients

        Returns
        -------
        np.ndarray wavelet convolved data

        """

        return filter_convolve_stack(data, self.filters, filter_rot=True)


class LinearCombo(object):
    """Linear combination class

    This class defines a combination of linear transform operators

    Parameters
    ----------
    operators : list
        List of linear operator class instances
    weights : list, optional
        List of weights for combining the linear adjoint operator results

    """

    def __init__(self, operators, weights=None):

        self.operators = operators
        self.weights = weights
        self.l1norm = np.array([operator.l1norm for operator in
                                self.operators])

    def op(self, data):
        """Operator

        This method returns the input data operated on by all of the operators

        Parameters
        ----------
        data : np.ndarray
            Input data array, a 2D image

        Returns
        -------
        np.ndarray linear operation results

        """

        res = np.empty(len(self.operators), dtype=np.ndarray)

        for i in range(len(self.operators)):
            res[i] = self.operators[i].op(data)

        return res

    def adj_op(self, data):
        """Adjoint operator

        This method returns the combination of the result of all of the
        adjoint operators. If weights are provided the comibination is the sum
        of the weighted results, otherwise the combination is the mean.

        Parameters
        ----------
        data : np.ndarray
            Input data array, an array of coefficients

        Returns
        -------
        np.ndarray adjoint operation results

        """

        if isinstance(self.weights, type(None)):

            return np.mean([operator.adj_op(x) for x, operator in
                           zip(data, self.operators)], axis=0)

        else:

            return np.sum([weight * operator.adj_op(x) for x, operator,
                          weight in zip(data, self.operators, weights)],
                          axis=0)



# -*- coding: utf-8 -*-

"""REWEIGHTING CLASSES

This module contains classes for reweighting optimisation implementations

:Author: Samuel Farrens <samuel.farrens@gmail.com>

:Version: 1.3

:Date: 20/10/2017

"""

from __future__ import division
import numpy as np


class cwbReweight(object):
    r"""Candes, Wakin and Boyd reweighting class

    This class implements the reweighting scheme described in [CWB2007]_

    Parameters
    ----------
    weights : np.ndarray
        Array of weights
    thresh_factor : float
        Threshold factor

    """

    def __init__(self, weights, thresh_factor=1):

        self.weights = weights
        self.original_weights = np.copy(self.weights)
        self.thresh_factor = thresh_factor

    def reweight(self, data):
        r"""Reweight

        This method implements the reweighting from section 4 in [CWB2007]_

        Notes
        -----

        Reweighting implemented as:

        .. math::

            w = w \left( \frac{1}{1 + \frac{|x^w|}{n \sigma}} \right)

        """

        self.weights *= (1.0 / (1.0 + np.abs(data) / (self.thresh_factor *
                         self.original_weights)))



# -*- coding: utf-8 -*-

"""PLOTTING ROUTINES

This module contains methods for making plots.

:Author: Samuel Farrens <samuel.farrens@gmail.com>

:Version: 1.1

:Date: 05/01/2017

"""

from __future__ import print_function
import numpy as np
import matplotlib.pyplot as plt


def plotCost(cost_list, output=None):
    """Plot cost function

    Plot the final cost function

    Parameters
    ----------
    cost_list : list
        List of cost function values
    output : str, optional
        Output file name

    """

    if isinstance(output, type(None)):
        file_name = 'cost_function.png'
    else:
        file_name = output + '_cost_function.png'

    plt.figure()
    plt.plot(np.log10(cost_list), 'r-')
    plt.title('Cost Function')
    plt.xlabel('Iteration')
    plt.ylabel('$\log_{10}$ Cost')
    plt.savefig(file_name)
    plt.close()

    print(' - Saving cost function data to:', file_name)



# -*- coding: utf-8 -*-

"""SVD ROUTINES

This module contains methods for thresholding singular values.

:Author: Samuel Farrens <samuel.farrens@gmail.com>

:Version: 1.2

:Date: 20/10/2017

"""

from __future__ import division
from builtins import zip
import numpy as np
from scipy.linalg import svd, diagsvd
from sf_tools.image.convolve import convolve
from sf_tools.signal.noise import thresh
from sf_tools.base.transform import cube2matrix, matrix2cube
from sf_tools.interface.errors import warn


def find_n_pc(u, factor=0.5):
    """Find number of principal components

    This method finds the minimum number of principal components required

    Parameters
    ----------
    u : np.ndarray
        Left singular vector
    factor : float, optional
        Factor for testing the auto correlation (default is '0.5')

    Returns
    -------
    int number of principal components

    """

    # Get the shape of the galaxy images.
    gal_shape = np.repeat(np.int(np.sqrt(u.shape[0])), 2)

    # Find the auto correlation of the left singular vector.
    u_auto = [convolve(a.reshape(gal_shape), np.rot90(a.reshape(gal_shape), 2))
              for a in u.T]

    # Return the required number of principal components.
    return np.sum(((a[list(zip(gal_shape // 2))] ** 2 <= factor *
                  np.sum(a ** 2)) for a in u_auto))


def svd_thresh(data, threshold=None, n_pc=None, thresh_type='hard'):
    """Threshold the singular values

    This method thresholds the input data using singular value decomposition

    Parameters
    ----------
    data : np.ndarray
        Input data array
    threshold : float, optional
        Threshold value
    n_pc : int or str, optional
        Number of principal components, specify an integer value or 'all'
    threshold_type : str {'hard', 'soft'}
        Type of noise to be added (default is 'hard')

    Returns
    -------
    np.ndarray thresholded data

    Raises
    ------
    ValueError
        For invalid string entry for n_pc

    """

    if isinstance(n_pc, str) and n_pc != 'all':
        raise ValueError('Invalid value for "n_pc", specify an integer value '
                         'or "all"')

    # Get SVD of input data.
    u, s, v = svd(data, check_finite=False, lapack_driver='gesvd')

    # Find the threshold if not provided.
    if isinstance(threshold, type(None)):

        # Find the required number of principal components if not specified.
        if isinstance(n_pc, type(None)):
            n_pc = find_n_pc(u, factor=0.1)

        # If the number of PCs is too large use all of the singular values.
        if n_pc >= s.size or n_pc == 'all':
            n_pc = s.size - 1
            warn('Using all singular values.')

        threshold = s[n_pc]

    # Remove noise from singular values.
    s_new = thresh(s, threshold, thresh_type)

    if np.all(s_new == s):
        warn('No change to singular values.')

    # Reshape the singular values to the shape of the input image.
    s_new = diagsvd(s_new, *data.shape)

    # Return the thresholded image.
    return np.dot(u, np.dot(s_new, v))


def svd_thresh_coef(data, operator, threshold, thresh_type='hard'):
    """Threshold the singular values coefficients

    This method thresholds the input data using singular value decomposition

    Parameters
    ----------
    data : np.ndarray
        Input data array
    operator : class
        Operator class instance
    threshold : float, optional
        Threshold value
    threshold_type : str {'hard', 'soft'}
        Type of noise to be added (default is 'hard')

    Returns
    -------
    np.ndarray thresholded data

    Raises
    ------
    ValueError
        For invalid string entry for n_pc

    """

    # Convert data cube to matrix.
    data_matrix = cube2matrix(data)

    # Get SVD of data matrix.
    u, s, v = np.linalg.svd(data_matrix, full_matrices=False)

    # Compute coefficients.
    a = np.dot(np.diag(s), v)

    # Compute threshold matrix.
    u_cube = matrix2cube(u, data.shape[1:])
    ti = np.array([np.linalg.norm(x) for x in operator(u_cube)])
    ti = np.repeat(ti, a.shape[1]).reshape(a.shape)
    threshold *= ti

    # Remove noise from coefficients.
    a_new = thresh(a, threshold, thresh_type)

    # Return the thresholded image.
    return np.dot(u, a_new)



# -*- coding: utf-8 -*-

"""MISCELLANOUS MATH ROUTINES

This module contains methods for various mathematical operations.

:Author: Samuel Farrens <samuel.farrens@gmail.com>

:Version: 1.1

:Date: 20/10/2017

"""

from __future__ import division
import numpy as np


def factor(n):
    """Factors of n

    This method finds factors of a number (n).

    Parameters
    ----------
    n : int
        Whole number

    Returns
    -------
    np.ndarray factors of n

    """

    factors = set()

    for x in range(1, int(np.sqrt(n)) + 1):
        if n % x == 0:
            factors.add(x)
            factors.add(n // x)

    return np.array(sorted(factors))


def mfactor(n):
    """Middle factors of n

    This method finds the middle factor(s) of a number (n).

    Parameters
    ----------
    n : int
        Whole number

    Returns
    -------
    np.ndarray middle factors of n

    """

    f = factor(n)

    if f.size % 2:
        return np.repeat(f[f.size // 2], 2)

    else:
        return f[f.size // 2 - 1:f.size // 2 + 1]


def k_val(n, L):
    """Spatial frequency

    This method returns k-values in the range L.

    Parameters
    ----------
    n : float
        Number
    limit : float
        Limit

    Returns
    -------
    float k value

    TODO
    ----
    Add equation and example

    """

    return ((2.0 * np.pi / limit) * np.array(range(n / 2.0) +
            range(-n / 2.0, 0.0)))


def fourier_derivative(func, k, order):
    """Fourier derivative

    This method returns the derivative of the specified function to the given
    order.

    Parameters
    ----------
    func : function
        Function
    k : float
        k-value
    order : int
        Oder of derivative

    Returns
    -------
        Float derivative

    TODO
    ----
    Add equation and example

    """

    return np.real(np.fft.ifft((1.j * k) ** order * np.fft.fft(func)))



# -*- coding: utf-8 -*-

r"""OPTIMISATION CLASSES

This module contains classes for optimisation algoritms.

:Author: Samuel Farrens <samuel.farrens@gmail.com>

:Version: 1.3

:Date: 20/10/2017

NOTES
-----

Input classes must have the following properties:

    * **Gradient Operators**

    Must have the following methods:

        * ``get_grad()`` - calculate the gradient

    Must have the following variables:

        * ``grad`` - the gradient
        * ``inv_spec_rad`` - inverse spectral radius :math:`\frac{1}{\rho}`

    * **Linear Operators**

    Must have the following methods:

        * ``op()`` - operator
        * ``adj_op()`` - adjoint operator

    Must have the following variables:

        * ``l1norm`` - the l1 norm of the operator

    * **Proximity Operators**

    Must have the following methods:

        * ``op()`` - operator

The following notation is used to implement the algorithms:

    * x_old is used in place of :math:`x_{n}`.
    * x_new is used in place of :math:`x_{n+1}`.
    * x_prox is used in place of :math:`\tilde{x}_{n+1}`.
    * x_temp is used for intermediate operations.

"""

from __future__ import division, print_function
from builtins import range, zip
import numpy as np


class FISTA(object):
    r"""FISTA

    This class is inhereited by optimisation classes to speed up convergence

    Parameters
    ----------
    lambda_init : float, optional
        Initial value of the relaxation parameter
    active : bool, optional
        Option to activate FISTA convergence speed-up (default is ``True``)

    """

    def __init__(self, lambda_init=None, active=True):

        self.lambda_now = lambda_init
        self.t_now = 1.0
        self.t_prev = 1.0
        self.use_speed_up = active

    def speed_switch(self, turn_on=True):
        r"""Speed swicth

        This method turns on or off the speed-up

        Parameters
        ----------
        turn_on : bool
            Option to turn on speed-up (default is ``True``)

        """

        self.use_speed_up = turn_on

    def update_lambda(self):
        r"""Update lambda

        This method updates the value of lambda

        Notes
        -----
        Implements steps 3 and 4 from algoritm 10.7 in [B2011]_

        """

        self.t_prev = self.t_now
        self.t_now = (1 + np.sqrt(4 * self.t_prev ** 2 + 1)) * 0.5
        self.lambda_now = 1 + (self.t_prev - 1) / self.t_now

    def speed_up(self):
        r"""speed-up

        This method returns the update if the speed-up is active

        """

        if self.use_speed_up:
            self.update_lambda()


class ForwardBackward(FISTA):
    r"""Forward-Backward optimisation

    This class implements standard forward-backward optimisation with an the
    option to use the FISTA speed-up

    Parameters
    ----------
    x : np.ndarray
        Initial guess for the primal variable
    grad : class
        Gradient operator class
    prox : class
        Proximity operator class
    cost : class, optional
        Cost function class
    lambda_init : float, optional
        Initial value of the relaxation parameter
    lambda_update : function, optional
        Relaxation parameter update method
    use_fista : bool, optional
        Option to use FISTA (default is ``True``)
    auto_iterate : bool, optional
        Option to automatically begin iterations upon initialisation (default
        is 'True')

    """

    def __init__(self, x, grad, prox, cost=None, lambda_init=None,
                 lambda_update=None, use_fista=True, auto_iterate=True):

        FISTA.__init__(self, lambda_init, use_fista)
        self.x_old = x
        self.z_old = np.copy(self.x_old)
        self.grad = grad
        self.prox = prox
        self.cost_func = cost
        self.lambda_update = lambda_update
        self.converge = False
        if auto_iterate:
            self.iterate()

    def update(self):
        r"""Update

        This method updates the current reconstruction

        Notes
        -----
        Implements algorithm 10.7 (or 10.5) from [B2011]_

        """

        # Step 1 from alg.10.7.
        self.grad.get_grad(self.z_old)
        y_old = self.z_old - self.grad.inv_spec_rad * self.grad.grad

        # Step 2 from alg.10.7.
        self.x_new = self.prox.op(y_old)

        # Steps 3 and 4 from alg.10.7.
        self.speed_up()

        # Step 5 from alg.10.7.
        self.z_new = self.x_old + self.lambda_now * (self.x_new - self.x_old)

        # Test primal variable for convergence.
        if np.sum(np.abs(self.z_old - self.z_new)) <= 1e-6:
            print(' - converged!')
            self.converge = True

        # Update old values for next iteration.
        np.copyto(self.x_old, self.x_new)
        np.copyto(self.z_old, self.z_new)

        # Update parameter values for next iteration.
        if not isinstance(self.lambda_update, type(None)):
            self.lambda_now = self.lambda_update(self.lambda_now)

        # Test cost function for convergence.
        if not isinstance(self.cost_func, type(None)):
            self.converge = self.cost_func.get_cost(self.z_new)

    def iterate(self, max_iter=150):
        r"""Iterate

        This method calls update until either convergence criteria is met or
        the maximum number of iterations is reached

        Parameters
        ----------
        max_iter : int, optional
            Maximum number of iterations (default is ``150``)

        """

        for i in range(max_iter):
            self.update()

            if self.converge:
                print(' - Converged!')
                break

        self.x_final = self.z_new


class GenForwardBackward(object):
    r"""Generalized Forward-Backward optimisation

    This class implements algorithm 1 from [R2012]_

    Parameters
    ----------
    x : np.ndarray
        Initial guess for the primal variable
    grad : class
        Gradient operator class
    prox_list : list
        List of proximity operator classes
    cost : class, optional
        Cost function class
    lambda_init : float, optional
        Initial value of the relaxation parameter
    lambda_update : function, optional
        Relaxation parameter update method
    weights : np.ndarray, optional
        Proximity operator weights
    auto_iterate : bool, optional
        Option to automatically begin iterations upon initialisation (default
        is 'True')

    """

    def __init__(self, x, grad, prox_list, cost=None, lambda_init=1.0,
                 lambda_update=None, weights=None, auto_iterate=True):

        self.x_old = x
        self.grad = grad
        self.prox_list = np.array(prox_list)
        self.cost_func = cost
        self.lambda_init = lambda_init
        self.lambda_update = lambda_update

        if isinstance(weights, type(None)):
            self.weights = np.repeat(1.0 / self.prox_list.size,
                                     self.prox_list.size)
        else:
            self.weights = np.array(weights)

        # Check weights.
        if np.sum(self.weights) != 1.0:
            raise ValueError('Proximity operator weights must sum to 1.0.'
                             'Current sum of weights = ' +
                             str(np.sum(self.weights)))

        self.z = np.array([self.x_old for i in range(self.prox_list.size)])

        self.converge = False
        if auto_iterate:
            self.iterate()

    def update(self):
        r"""Update

        This method updates the current reconstruction

        Notes
        -----
        Implements algorithm 1 from [R2012]_

        """

        # Calculate gradient for current iteration.
        self.grad.get_grad(self.x_old)

        # Update z values.
        for i in range(self.prox_list.size):
            z_temp = (2 * self.x_old - self.z[i] - self.grad.inv_spec_rad *
                      self.grad.grad)
            z_prox = self.prox_list[i].op(z_temp,
                                          extra_factor=self.grad.inv_spec_rad /
                                          self.weights[i])
            self.z[i] += self.lambda_init * (z_prox - self.x_old)

        # Update current reconstruction.
        self.x_new = np.sum((z_i * w_i for z_i, w_i in
                            zip(self.z, self.weights)), axis=0)

        # Update old values for next iteration.
        np.copyto(self.x_old, self.x_new)

        # Update parameter values for next iteration.
        if not isinstance(self.lambda_update, type(None)):
            self.lambda_now = self.lambda_update(self.lambda_now)

        # Test cost function for convergence.
        if not isinstance(self.cost_func, type(None)):
            self.converge = self.cost_func.get_cost(self.x_new)

    def iterate(self, max_iter=150):
        r"""Iterate

        This method calls update until either convergence criteria is met or
        the maximum number of iterations is reached

        Parameters
        ----------
        max_iter : int, optional
            Maximum number of iterations (default is ``150``)

        """

        for i in range(max_iter):
            self.update()

            if self.converge:
                print(' - Converged!')
                break

        self.x_final = self.x_new


class Condat(object):
    r"""Condat optimisation

    This class implements algorithm 10.7 from [Con2013]_

    Parameters
    ----------
    x : np.ndarray
        Initial guess for the primal variable
    y : np.ndarray
        Initial guess for the dual variable
    grad : class
        Gradient operator class
    prox : class
        Proximity primal operator class
    prox_dual : class
        Proximity dual operator class
    linear : class
        Linear operator class
    cost : class
        Cost function class
    rho : float
        Relaxation parameter
    sigma : float
        Proximal dual parameter
    tau : float
        Proximal primal paramater
    rho_update : function, optional
        Relaxation parameter update method
    sigma_update : function, optional
        Proximal dual parameter update method
    tau_update : function, optional
        Proximal primal parameter update method
    auto_iterate : bool, optional
        Option to automatically begin iterations upon initialisation (default
        is 'True')

    """

    def __init__(self, x, y, grad, prox, prox_dual, linear, cost,
                 rho,  sigma, tau, rho_update=None, sigma_update=None,
                 tau_update=None, auto_iterate=True):

        self.x_old = x
        self.y_old = y
        self.grad = grad
        self.prox = prox
        self.prox_dual = prox_dual
        self.linear = linear
        self.cost_func = cost
        self.rho = rho
        self.sigma = sigma
        self.tau = tau
        self.rho_update = rho_update
        self.sigma_update = sigma_update
        self.tau_update = tau_update
        self.converge = False
        if auto_iterate:
            self.iterate()

    def update_param(self):
        r"""Update parameters

        This method updates the values of ``rho``, ``sigma`` and ``tau`` with
        the methods provided

        """

        # Update relaxation parameter.
        if not isinstance(self.rho_update, type(None)):
            self.rho = self.rho_update(self.rho)

        # Update proximal dual parameter.
        if not isinstance(self.sigma_update, type(None)):
            self.sigma = self.sigma_update(self.sigma)

        # Update proximal primal parameter.
        if not isinstance(self.tau_update, type(None)):
            self.tau = self.tau_update(self.tau)

    def update(self):
        r"""Update

        This method updates the current reconstruction

        Notes
        -----
        Implements equation 9 (algorithm 3.1) from [Con2013]_

        - primal proximity operator set up for positivity constraint

        """

        # Step 1 from eq.9.
        self.grad.get_grad(self.x_old)

        x_prox = self.prox.op(self.x_old - self.tau * self.grad.grad -
                              self.tau * self.linear.adj_op(self.y_old))

        # Step 2 from eq.9.
        y_temp = (self.y_old + self.sigma *
                  self.linear.op(2 * x_prox - self.x_old))

        y_prox = (y_temp - self.sigma * self.prox_dual.op(y_temp / self.sigma,
                  extra_factor=(1.0 / self.sigma)))

        # Step 3 from eq.9.
        self.x_new = self.rho * x_prox + (1 - self.rho) * self.x_old
        self.y_new = self.rho * y_prox + (1 - self.rho) * self.y_old

        del x_prox, y_prox, y_temp

        # Update old values for next iteration.
        np.copyto(self.x_old, self.x_new)
        np.copyto(self.y_old, self.y_new)

        # Update parameter values for next iteration.
        self.update_param()

        # Test cost function for convergence.
        self.converge = self.cost_func.get_cost(self.x_new)

    def iterate(self, max_iter=150):
        r"""Iterate

        This method calls update until either convergence criteria is met or
        the maximum number of iterations is reached

        Parameters
        ----------
        max_iter : int, optional
            Maximum number of iterations (default is ``150``)

        """

        for i in range(max_iter):
            self.update()

            if self.converge:
                print(' - Converged!')
                break

        self.x_final = self.x_new
        self.y_final = self.y_new



from setuptools import setup, find_packages

setup(
    name='sf_tools',
    author='sfarrens',
    author_email='samuel.farrens@gmail.com',
    version='1.1',
    url='https://github.com/sfarrens/sf_tools',
    download_url='https://github.com/sfarrens/sf_tools/archive/0.1.tar.gz',
    packages=find_packages(),
    license='MIT',
    long_description=open('README.txt').read(),
)



# -*- coding: utf-8 -*-

"""CONVOLUTION ROUTINES

This module contains methods for convolution.

:Author: Samuel Farrens <samuel.farrens@gmail.com>

:Version: 1.2

:Date: 23/10/2017

"""

from __future__ import division
from builtins import zip
import numpy as np
from scipy.signal import fftconvolve
from scipy.fftpack import fftn, ifftn, fftshift, ifftshift
from astropy.convolution import convolve_fft
from sf_tools.base.np_adjust import rotate, rotate_stack


def convolve_np(image, kernel):
    """Convolve with Numpy FFT

    This method convolves the input image with the input kernel

    Parameters
    ----------
    image : np.ndarray
        2D image array
    kernel : np.ndarray
        2D kernel array

    Returns
    -------
    np.ndarray 2D convolved image array

    """

    x = np.fft.fftshift(np.fft.fftn(image))
    y = np.fft.fftshift(np.fft.fftn(kernel))

    return np.real(np.fft.fftshift(np.fft.ifftn(np.fft.ifftshift(x * y))))


def deconvolve_np(image, kernel):
    """Deconvolve with Numpy FFT

    This method deconvolves the input image with the input kernel

    Parameters
    ----------
    image : np.ndarray
        2D image array
    kernel : np.ndarray
        2D kernel array

    Returns
    -------
    np.ndarray 2D deconvolved image array

    """

    x = np.fft.fftshift(np.fft.fftn(image))
    y = np.fft.fftshift(np.fft.fftn(kernel))

    return np.real(np.fft.fftshift(np.fft.ifftn(np.fft.ifftshift(x / y))))


def convolve(data, kernel, method='astropy'):
    """Convolve data with kernel

    This method convolves the input data with a given kernel using FFT and
    is the default convolution used for all routines

    Parameters
    ----------
    data : np.ndarray
        Input data array, normally a 2D image
    kernel : np.ndarray
        Input kernel array, normally a 2D kernel
    method : str {'astropy', 'scipy'}, optional
        Convolution method (default is 'astropy')

        'astropy':
            Uses the astropy.convolution.convolve_fft method provided in
            Astropy (http://www.astropy.org/)

        'scipy':
            Uses the scipy.signal.fftconvolve method provided in SciPy
            (https://www.scipy.org/)

    Returns
    -------
    np.ndarray convolved data

    Raises
    ------
    ValueError
        If `data` and `kernel` do not have the same number of dimensions
    ValueError
        If `method` is not 'astropy' or 'scipy'

    """

    if data.ndim != kernel.ndim:
        raise ValueError('Data and kernel must have the same dimensions.')

    if method not in ('astropy', 'scipy'):
        raise ValueError('Invalid method. Options are "astropy" or "scipy".')

    if method == 'astropy':
        return convolve_fft(data, kernel, boundary='wrap', crop=False,
                            nan_treatment='fill', normalize_kernel=False)

    elif method == 'scipy':
        return fftconvolve(data, kernel, mode='same')


def convolve_stack(data, kernel, rot_kernel=False, method='astropy'):
    """Convolve stack of data with stack of kernels

    This method convolves the input data with a given kernel using FFT and
    is the default convolution used for all routines

    Parameters
    ----------
    data : np.ndarray
        Input data array, normally a 2D image
    kernel : np.ndarray
        Input kernel array, normally a 2D kernel
    rot_kernel : bool
        Option to rotate kernels by 180 degrees
    method : str {'astropy', 'scipy'}, optional
        Convolution method (default is 'astropy')

    Returns
    -------
    np.ndarray convolved data

    """

    if rot_kernel:
        kernel = rotate_stack(kernel)

    return np.array([convolve(data_i, kernel_i, method=method) for data_i,
                    kernel_i in zip(data, kernel)])


def psf_convolve(data, psf, psf_rot=False, psf_type='fixed', method='astropy'):
    """Convolve data with PSF

    This method convolves an image with a PSF

    Parameters
    ----------
    data : np.ndarray
        Input data array, normally an array of 2D images
    psf : np.ndarray
        Input PSF array, normally either a single 2D PSF or an array of 2D
        PSFs
    psf_rot: bool
        Option to rotate PSF by 180 degrees
    psf_type : str {'fixed', 'obj_var'}, optional
        PSF type (default is 'fixed')
    method : str {'astropy', 'scipy'}, optional
        Convolution method (default is 'astropy')

        'fixed':
            The PSF is fixed, i.e. it is the same for each image

        'obj_var':
            The PSF is object variant, i.e. it is different for each image

    Returns
    -------
    np.ndarray convolved data

    Raises
    ------
    ValueError
        If `psf_type` is not 'fixed' or 'obj_var'

    """

    if psf_type not in ('fixed', 'obj_var'):
        raise ValueError('Invalid PSF type. Options are "fixed" or "obj_var"')

    if psf_rot and psf_type == 'fixed':
        psf = rotate(psf)

    elif psf_rot:
        psf = rotate_stack(psf)

    if psf_type == 'fixed':
        return np.array([convolve(data_i, psf, method=method) for data_i in
                        data])

    elif psf_type == 'obj_var':

        return convolve_stack(data, psf)


def pseudo_inverse(image, kernel, weight=None):
    """Pseudo inverse

    This method calculates the pseudo inverse of the input image for the given
    kernel using FFT

    Parameters
    ----------
    image : np.ndarray
        Input image, 2D array
    kernel : np.ndarray
        Input kernel, 2D array
    weight : np.ndarray, optional
        Optional weights, 2D array

    Returns
    -------
    np.ndarray result of the pseudo inverse

    """

    y_hat = fftshift(fftn(image))
    h_hat = fftshift(fftn(kernel))
    h_hat_star = np.conj(h_hat)

    res = ((h_hat_star * y_hat) / (h_hat_star * h_hat))

    if not isinstance(weight, type(None)):
        res *= weight

    return np.real(fftshift(ifftn(ifftshift(res))))



# -*- coding: utf-8 -*-

"""ANGLE HANDLING ROUTINES

This module contains methods for handing angles and trigonometry.

:Author: Samuel Farrens <samuel.farrens@gmail.com>

:Version: 1.1

:Date: 20/10/2017

"""

from __future__ import division
import numpy as np


def deg2rad(angle):
    r"""Degrees to radians

    This method converts the angle from degrees to radians.

    Parameters
    ----------
    angle : float or np.ndarray
        Input angle in degrees

    Returns
    -------
    float angle in radians or np.ndarray of angles

    Examples
    --------
    >>> from sf_tools.math.angle import deg2rad
    >>> deg2rad(45.)
    0.7853981633974483

    Notes
    -----
    Implements the following equation:

    .. math::
        \mathrm{radians} = \mathrm{degrees} \times \frac{\pi}{180}

    """

    return angle * np.pi / 180.0


def rad2deg(angle):
    r"""Radians to degrees

    This method converts the angle from radians to degrees.

    Parameters
    ----------
    angle : float or np.ndarray
        Input angle in radians

    Returns
    -------
    float angle in degrees or np.ndarray of angles

    Examples
    --------
    >>> from sf_tools.math.angle import deg2rad
    >>> rad2deg(1.)
    57.29577951308232

    Notes
    -----
    Implements the following equation:

    .. math::
        \mathrm{degrees} = \mathrm{radians} \times \frac{180}{\pi}

    """

    return angle * 180.0 / np.pi


def ang_sep(point1, point2):
    r"""Angular separation

    This method calculates the angular separation in degrees between two
    points.

    Parameters
    ----------
    point1 : tuple
        Angular position of point 1 in degrees
    point1 : tuple
        Angular position of point 2 in degrees

    Returns
    -------
    float angular separation in degrees

    Examples
    --------
    >>> from sf_tools.math.angle import ang_sep
    >>> ang_sep((30.0, 0.0), (47.0, 10.0))
    19.647958606833164

    Notes
    -----
    Implements the following equation:

    .. math::
        \theta = \cos^{-1}[\sin(\delta_1)\sin(\delta_2)+
        \cos(\delta_1)\cos(\delta_2)\cos(\alpha_1-\alpha_2)]

    See https://en.wikipedia.org/wiki/Angular_distance

    """

    dist = np.around(np.sin(deg2rad(point1[1])) * np.sin(deg2rad(point2[1])) +
                     np.cos(deg2rad(point1[1])) * np.cos(deg2rad(point2[1])) *
                     np.cos(deg2rad(point1[0]) - deg2rad(point2[0])), 10)

    return rad2deg(np.array(np.arccos(dist)))



# -*- coding: utf-8 -*-

"""QUALITY ASSESSMENT ROUTINES

This module contains methods and classes for assessing the quality of image
reconstructions.

:Author: Samuel Farrens <samuel.farrens@gmail.com>

:Version: 1.2

:Date: 20/10/2017

Notes
-----
Some of the methods in this module are based on work by Fred Ngole.

"""

from __future__ import division
import numpy as np
from sf_tools.image.shape import ellipticity_atoms


def nmse(image1, image2, metric=np.mean):
    r"""Normalised Mean Square Error

    This method computes the NMSE of two input images, or the result of the
    input metric on a stack of input images.

    Parameters
    ----------
    image1 : np.ndarray
        First image (or stack of images) to be analysed (original image)
    image2 : np.ndarray
        Second image (or stack of images) to be analysed (reconstructed image)
    metric : function
        Metric to be apllied to NMSE results (default is 'np.mean')

    Returns
    -------
    float NMSE value or metric value(s)

    Raises
    ------
    ValueError
        For invalid input data dimensions

    See Also
    --------
    e_error : ellipticity error

    Notes
    -----
    This method implements the following equation:

        - Equations from [NS2016]_ sec 4.1:

        .. math::

            \text{NMSE} = \frac{1}{D}\sum_{i=1}^D
                \frac{\|\hat{\text{Im}}_i - \text{Im}_i\|_2^2}
                {\|\text{Im}_i\|_2^2}

    Examples
    --------
    >>> from image.quality import nmse

    """

    if image1.ndim != image2.ndim:
        raise ValueError('Input images must have the same dimensions')

    if image1.ndim not in (2, 3):
        raise ValueError('Input data must be single image or stack of images')

    if image1.ndim == 2:

        return (np.linalg.norm(image2 - image1) ** 2 /
                np.linalg.norm(image1) ** 2)

    else:

        res = (np.array([np.linalg.norm(x) ** 2 for x in (image2 - image1)]) /
               np.array([np.linalg.norm(x) ** 2 for x in image1]))

        return metric(res)


def e_error(image1, image2, metric=np.mean):
    r"""Normalised Mean Square Error

    This method computes the ellipticity error of two input images, or the
    result of the input metric on the ellipticity values.

    Parameters
    ----------
    image1 : np.ndarray
        First image to be analysed (original image)
    image2 : np.ndarray
        Second image to be analysed (reconstructed image)
    metric : function
        Metric to be apllied to ellipticity error results (default is
        'np.mean')

    Returns
    -------
    float ellipticity error value or metric value(s)

    Raises
    ------
    ValueError
        For invalid input data dimensions

    See Also
    --------
    nmse : nmse error

    Notes
    -----
    This method implements the following equation:

        - Equations from [NS2016]_ sec 4.1:

        .. math::

            \text{E}_\gamma = \frac{1}{D}\sum_{i=1}^D
                \|\gamma(\text{Im}_i) - \gamma(\hat{\text{Im}}_i)\|_2

    Examples
    --------
    >>> from image.quality import e_error

    """

    if image1.ndim != image2.ndim:
        raise ValueError('Input images must have the same dimensions')

    if image1.ndim not in (2, 3):
        raise ValueError('Input data must be single image or stack of images')

    if image1.ndim == 2:

        return np.linalg.norm(ellipticity_atoms(image1) -
                              ellipticity_atoms(image2))

    else:

        diff = (np.array([ellipticity_atoms(x) for x in image1]) -
                np.array([ellipticity_atoms(x) for x in image2]))

        return metric([np.linalg.norm(x) for x in diff])



# -*- coding: utf-8 -*-

"""NOISE ROUTINES

This module contains methods for adding and removing noise from data.

:Author: Samuel Farrens <samuel.farrens@gmail.com>

:Version: 1.2

:Date: 04/01/2017

"""

from builtins import zip
import numpy as np


def add_noise(data, sigma=1.0, noise_type='gauss'):
    r"""Add noise to data

    This method adds Gaussian or Poisson noise to the input data

    Parameters
    ----------
    data : np.ndarray, list or tuple
        Input data array
    sigma : float or list, optional
        Standard deviation of the noise to be added ('gauss' only)
    noise_type : str {'gauss', 'poisson'}
        Type of noise to be added (default is 'gauss')

    Returns
    -------
    np.ndarray input data with added noise

    Raises
    ------
    ValueError
        If `noise_type` is not 'gauss' or 'poisson'
    ValueError
        If number of `sigma` values does not match the first dimension of the
        input data

    Examples
    --------
    >>> import numpy as np
    >>> from sf_tools.signal.noise import add_noise
    >>> x = np.arange(9).reshape(3, 3).astype(float)
    >>> x
    array([[ 0.,  1.,  2.],
           [ 3.,  4.,  5.],
           [ 6.,  7.,  8.]])
    >>> np.random.seed(1)
    >>> add_noise(x, noise_type='poisson')
    array([[  0.,   2.,   2.],
           [  4.,   5.,  10.],
           [ 11.,  15.,  18.]])

    >>> import numpy as np
    >>> from sf_tools.signal.noise import add_noise
    >>> x = np.zeros(5)
    >>> x
    array([ 0.,  0.,  0.,  0.,  0.])
    >>> np.random.seed(1)
    >>> add_noise(x, sigma=2.0)
    array([ 3.24869073, -1.22351283, -1.0563435 , -2.14593724,  1.73081526])

    """

    data = np.array(data)

    if noise_type not in ('gauss', 'poisson'):
        raise ValueError('Invalid noise type. Options are "gauss" or'
                         '"poisson"')

    if isinstance(sigma, (list, np.ndarray)):
        if len(sigma) != data.shape[0]:
            raise ValueError('Number of sigma values must match first '
                             'dimension of input data')

    if noise_type is 'gauss':
        random = np.random.randn(*data.shape)

    elif noise_type is 'poisson':
        random = np.random.poisson(np.abs(data))

    if isinstance(sigma, float):
        return data + sigma * random

    else:
        return data + np.array([s * r for s, r in zip(sigma, random)])


def thresh(data, threshold, threshold_type='hard'):
    r"""Threshold data

    This method perfoms hard or soft thresholding on the input data

    Parameters
    ----------
    data : np.ndarray, list or tuple
        Input data array
    threshold : float
        Threshold level
    threshold_type : str {'hard', 'soft'}
        Type of noise to be added (default is 'hard')

    Returns
    -------
    np.ndarray thresholded data

    Raises
    ------
    ValueError
        If `threshold_type` is not 'hard' or 'soft'


    Notes
    -----
    Implements one of the following two equations:

    * Hard Threshold
        .. math::
            \mathrm{HT}_\lambda(x) =
            \begin{cases}
            x & \text{if } |x|\geq\lambda \\
            0 & \text{otherwise}
            \end{cases}

    * Soft Threshold
        .. math::
            \mathrm{ST}_\lambda(x) =
            \begin{cases}
            x-\lambda\text{sign}(x) & \text{if } |x|\geq\lambda \\
            0 & \text{otherwise}
            \end{cases}

    Examples
    --------
    >>> import numpy as np
    >>> from sf_tools.signal.noise import thresh
    >>> np.random.seed(1)
    >>> x = np.random.randint(-9, 9, 10)
    >>> x
    array([-4,  2,  3, -1,  0,  2, -4,  6, -9,  7])
    >>> thresh(x, 4)
    array([-4,  0,  0,  0,  0,  0, -4,  6, -9,  7])

    >>> import numpy as np
    >>> from sf_tools.signal.noise import thresh
    >>> np.random.seed(1)
    >>> x = np.random.ranf((3, 3))
    >>> x
    array([[  4.17022005e-01,   7.20324493e-01,   1.14374817e-04],
           [  3.02332573e-01,   1.46755891e-01,   9.23385948e-02],
           [  1.86260211e-01,   3.45560727e-01,   3.96767474e-01]])
    >>> thresh(x, 0.2, threshold_type='soft')
    array([[ 0.217022  ,  0.52032449, -0.        ],
           [ 0.10233257, -0.        , -0.        ],
           [-0.        ,  0.14556073,  0.19676747]])

    """

    data = np.array(data)

    if threshold_type not in ('hard', 'soft'):
        raise ValueError('Invalid threshold type. Options are "hard" or'
                         '"soft"')

    if threshold_type == 'soft':
        return (np.sign(data) * (np.abs(data) - threshold) *
                (np.abs(data) >= threshold))

    else:
        return data * (np.abs(data) >= threshold)



# -*- coding: utf-8 -*-

"""TYPE HANDLING ROUTINES

This module contains methods for handing object types.

:Author: Samuel Farrens <samuel.farrens@gmail.com>

:Version: 1.0

:Date: 04/04/2017

"""


import numpy as np


def check_float(val):
    """Check if input value is a float or a np.ndarray of floats, if not
    convert.

    Parameters
    ----------
    val : any
        Input value

    Returns
    -------
    float or np.ndarray of floats

    """

    if type(val) is float:
        pass
    elif type(val) is int:
        val = float(val)
    elif type(val) is list or type(val) is tuple:
        val = np.array(val, dtype=float)
    elif type(val) is np.ndarray and val.dtype is not 'float64':
        val = val.astype(float)
    else:
        raise ValueError('Invalid input type.')

    return val


def check_int(val):
    """Check if input value is an int or a np.ndarray of ints, if not convert.

    Parameters
    ----------
    val : any
        Input value

    Returns
    -------
    int or np.ndarray of ints

    """

    if type(val) is int:
        pass
    elif type(val) is float:
        val = int(val)
    elif type(val) is list or type(val) is tuple:
        val = np.array(val, dtype=int)
    elif type(val) is np.ndarray and val.dtype is not 'int64':
        val = val.astype(int)
    else:
        raise ValueError('Invalid input type.')

    return val


def val2int(val):
    """Convert to int

    This method checks if input value is an int and if not converts it.

    Parameters
    ----------
    val : int, float, str, list, tuple or np.ndarray
        Input value

    Returns
    -------
    int interger value or np.ndarray array of ints

    Raises
    ------
    ValueError
        For invalid input type

    Examples
    --------
    >>> from sf_tools.base.types import val2int
    >>> x = np.arange(5).astype(float)
    >>> x
    array([ 0.,  1.,  2.,  3.,  4.])
    >>> val2int(x)
    array([0, 1, 2, 3, 4])

    """

    if isinstance(val, int):
        pass
    elif isinstance(val, (float, str)):
        val = int(float(val))
    elif isinstance(val, (list, tuple)):
        val = np.array(val, dtype=float).astype(int)
    elif isinstance(val, np.ndarray):
        if np.issubdtype(val.dtype, 'int64'):
            pass
        else:
            val = val.astype(float).astype(int)
    else:
        raise ValueError('Invalid input type.')

    return val


def val2float(val):
    """Convert to float

    This method checks if input value is a float and if not converts it.

    Parameters
    ----------
    val : int, float, str, list, tuple or np.ndarray
        Input value

    Returns
    -------
    float floating point value or np.ndarray array of floats

    Examples
    --------
    >>> from sf_tools.base.types import val2float
    >>> x = np.arange(5)
    >>> x
    array([0, 1, 2, 3, 4])
    >>> val2float(x)
    array([ 0.,  1.,  2.,  3.,  4.])

    """

    if isinstance(val, float):
        pass
    elif isinstance(val, (int, str)):
        val = float(val)
    elif isinstance(val, (list, tuple)):
        val = np.array(val, dtype=float)
    elif isinstance(val, np.ndarray):
        if np.issubdtype(val.dtype, 'float64'):
            pass
        else:
            val = val.astype(float)
    else:
        raise ValueError('Invalid input type.')

    return val


def val2str(val):
    """Convert to string

    This method checks if input value is a string and if not converts it.

    Parameters
    ----------
    val : int, float, str, list, tuple or np.ndarray
        Input value

    Returns
    -------
    str string or np.ndarray array of strings


    Examples
    --------
    >>> from sf_tools.base.types import val2str
    >>> x = np.arange(5)
    >>> x
    array([0, 1, 2, 3, 4])
    >>> val2str(x)
    array(['0', '1', '2', '3', '4'],
          dtype='|S21')

    """

    if isinstance(val, str):
        pass
    elif isinstance(val, (int, float)):
        val = str(val)
    elif isinstance(val, (list, tuple)):
        val = np.array(val, dtype=str)
    elif isinstance(val, np.ndarray):
        if np.issubdtype(val.dtype, 'S21'):
            pass
        else:
            val = val.astype(str)
    else:
        raise ValueError('Invalid input type.')

    return val


def nan2val(array, val=0.0):
    """Convert NAN to val

    This converts all NANs in an array to a specified value.

    Parameters
    ----------
    array : np.ndarray, list or tuple
        Input array
    val : int or float, optional
        Value to replace NANs. Default (val=0.0)

    Returns
    -------
    np.ndarray array without NANs

    NOTES
    -----
    Output data type defined by val type.

    Examples
    --------
    >>> from sf_tools.base.types import nan2val
    >>> x = [1., 2., np.nan, 4.]
    >>> x
    [1.0, 2.0, nan, 4.0]
    >>> nan2val(x, 3.)
    array([ 1.,  2.,  3.,  4.])

    """

    new_array = np.copy(array)
    new_array[np.isnan(new_array)] = val

    if isinstance(val, int):
        new_array = val2int(new_array)

    return new_array



"""TESTING ROUTINES

This module contains methods for testing signal and operator properties.

:Author: Samuel Farrens <samuel.farrens@gmail.com>

:Version: 1.1

:Date: 03/04/2017

"""

from __future__ import print_function
import numpy as np


def transpose_test(operator, operator_t, x_shape, x_args, y_shape=None,
                   y_args=None):
    """Transpose test

    This method tests two operators to see if they are the transpose of each
    other.

    Parameters
    ----------
    operator : function
        Operator function
    operator_t : function
        Transpose operator function
    x_shape : tuple
        Shape of operator input data
    x_args : tuple
        Arguments to be passed to operator
    y_shape : tuple, optional
        Shape of transpose operator input data
    y_args : tuple, optional
        Arguments to be passed to transpose operator

    """

    if isinstance(y_shape, type(None)):
        y_shape = x_shape

    if isinstance(y_args, type(None)):
        y_args = x_args

    # Generate random arrays.
    x = np.random.ranf(x_shape)
    y = np.random.ranf(y_shape)

    # Calculate <MX, Y>
    mx_y = np.sum(np.multiply(operator(x, *x_args), y))

    # Calculate <X, M.TY>
    x_mty = np.sum(np.multiply(x, operator_t(y, *y_args)))

    # Test the difference between the two.
    print(' - |<MX, Y> - <X, M.TY>| =', np.abs(mx_y - x_mty))



# -*- coding: utf-8 -*-

"""NUMPY ADJUSTMENT ROUTINES

This module contains methods for adjusting the default output for certain
Numpy functions.

:Author: Samuel Farrens <samuel.farrens@gmail.com>

:Version: 1.1

:Date: 03/04/2017

"""

from __future__ import division
import numpy as np


def rotate(data):
    """Rotate

    This method rotates an input numpy array by 180 degrees.

    Parameters
    ----------
    data : np.ndarray
        Input data array (at least 2D)

    Returns
    -------
    np.ndarray rotated data

    Notes
    -----
    Adjustment to numpy.rot90()

    Examples
    --------
    >>> from sf_tools.base.np_adjust import rotate
    >>> x = np.arange(9).reshape((3, 3))
    >>> x
    array([[0, 1, 2],
           [3, 4, 5],
           [6, 7, 8]])
    >>> rotate(x)
    array([[8, 7, 6],
           [5, 4, 3],
           [2, 1, 0]])

    """

    return np.rot90(data, 2)


def rotate_stack(data):
    """Rotate stack

    This method rotates each array in a stack of arrays by 180 degrees.

    Parameters
    ----------
    data : np.ndarray
        Input data array (at least 3D)

    Returns
    -------
    np.ndarray rotated data

    Examples
    --------
    >>> from sf_tools.base.np_adjust import rotate_stack
    >>> x = np.arange(18).reshape((2, 3, 3))
    >>> x
    array([[[ 0,  1,  2],
            [ 3,  4,  5],
            [ 6,  7,  8]],
           [[ 9, 10, 11],
            [12, 13, 14],
            [15, 16, 17
    >>> rotate_stack(x)
    array([[[ 8,  7,  6],
            [ 5,  4,  3],
            [ 2,  1,  0]],
           [[17, 16, 15],
            [14, 13, 12],
            [11, 10,  9]]])

    """

    return np.array([rotate(x) for x in data])


def pad2d(data, padding):
    """Pad array

    This method pads an input numpy array with zeros in all directions.

    Parameters
    ----------
    data : np.ndarray
        Input data array (at least 2D)
    shape : tuple
        Amount of padding in x and y directions, respectively

    Returns
    -------
    np.ndarray padded data

    Notes
    -----
    Adjustment to numpy.pad()

    Examples
    --------
    >>> from sf_tools.base.np_adjust import pad2d
    >>> x = np.arange(9).reshape((3, 3))
    >>> x
    array([[0, 1, 2],
           [3, 4, 5],
           [6, 7, 8]])
    >>> pad2d(x, (1, 1))
    array([[0, 0, 0, 0, 0],
           [0, 0, 1, 2, 0],
           [0, 3, 4, 5, 0],
           [0, 6, 7, 8, 0],
           [0, 0, 0, 0, 0]])

    """

    data = np.array(data)
    shape = np.array(data.shape)

    if shape.ndim == 1:
        shape = np.repeat(shape, 2)

    return np.pad(data, ((shape[0], shape[0]), (shape[1], shape[1])),
                  'constant')


def x_bins(vals):
    """X-range bins

    This method sets the bin values for a histogram.

    Parameters
    ----------
    vals : np.ndarray
        X-range bins from np.histogram()[1]

    Returns
    -------
    np.ndarray corrected x-range bin data

    Notes
    -----
    Adjustment to numpy.histogram()

    Examples
    --------
    >>> from sf_tools.base.np_adjust import x_bins
    >>> data = np.array([1, 2, 1, 3, 1, 1])
    >>> hist, bins = np.histogram(data)
    >>> hist
    array([4, 0, 0, 0, 0, 1, 0, 0, 0, 1])
    >>> bins
    array([ 1. ,  1.2,  1.4,  1.6,  1.8,  2. ,  2.2,  2.4,  2.6,  2.8,  3. ])
    >>> x_bins(bins)
    array([ 1.1,  1.3,  1.5,  1.7,  1.9,  2.1,  2.3,  2.5,  2.7,  2.9])

    """

    return (vals[:-1] + vals[1:]) / 2.0


def x_bins_step(vals):
    """X-range bins (step function)

    This method sets the bin values for a histogram plotted as a step funciton.

    Parameters
    ----------
    vals : np.ndarray
        X-range bins from np.histogram()[1]

    Returns
    -------
    np.ndarray corrected x-range bin data

    Notes
    -----
    Adjustment to numpy.histogram()

    Examples
    --------
    >>> from sf_tools.base.np_adjust import x_bins
    >>> data = np.array([1, 2, 1, 3, 1, 1])
    >>> hist, bins = np.histogram(data)
    >>> hist
    array([4, 0, 0, 0, 0, 1, 0, 0, 0, 1])
    >>> bins
    array([ 1. ,  1.2,  1.4,  1.6,  1.8,  2. ,  2.2,  2.4,  2.6,  2.8,  3. ])
    >>> x_bins_step(bins)
    array([ 1.2,  1.4,  1.6,  1.8,  2. ,  2.2,  2.4,  2.6,  2.8,  3. ])

    """

    return x_bins(vals) + (vals[1] - vals[0]) / 2.0


def ftr(data):
    """Fancy transpose right

    Apply fancy_transpose() to data with roll=1

    Parameters
    ----------
    data : np.ndarray
        Input data array

    Returns
    -------
    np.ndarray transposed data

    """

    return fancy_transpose(data)


def ftl(data):
    """Fancy transpose left

    Apply fancy_transpose() to data with roll=-1

    Parameters
    ----------
    data : np.ndarray
        Input data array

    Returns
    -------
    np.ndarray transposed data

    """

    return fancy_transpose(data, -1)


def fancy_transpose(data, roll=1):
    """Fancy transpose

    This method transposes a multidimensional matrix.

    Parameters
    ----------
    data : np.ndarray
        Input data array
    roll : int
        Roll direction and amount. Default (roll=1)

    Returns
    -------
    np.ndarray transposed data

    Notes
    -----
    Adjustment to numpy.transpose

    Examples
    --------
    >>> from sf_tools.base.np_adjust import fancy_transpose
    >>> x = np.arange(27).reshape(3, 3, 3)
    >>> x
    array([[[ 0,  1,  2],
            [ 3,  4,  5],
            [ 6,  7,  8]],
           [[ 9, 10, 11],
            [12, 13, 14],
            [15, 16, 17]],
           [[18, 19, 20],
            [21, 22, 23],
            [24, 25, 26]]])
    >>> fancy_transpose(x)
    array([[[ 0,  3,  6],
            [ 9, 12, 15],
            [18, 21, 24]],
           [[ 1,  4,  7],
            [10, 13, 16],
            [19, 22, 25]],
           [[ 2,  5,  8],
            [11, 14, 17],
            [20, 23, 26]]])
    >>> fancy_transpose(x, roll=-1)
    array([[[ 0,  9, 18],
            [ 1, 10, 19],
            [ 2, 11, 20]],
           [[ 3, 12, 21],
            [ 4, 13, 22],
            [ 5, 14, 23]],
           [[ 6, 15, 24],
            [ 7, 16, 25],
            [ 8, 17, 26]]])

    """

    axis_roll = np.roll(np.arange(data.ndim), roll)

    return np.transpose(data, axes=axis_roll)
